import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class PolarBearVidID(DatasetFactory):
    summary = summary['PolarBearVidID']
    url = 'https://zenodo.org/records/7564529/files/PolarBearVidID.zip?download=1'
    archive = 'PolarBearVidID.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        metadata = pd.read_csv(os.path.join(self.root, 'animal_db.csv'))
        data = utils.find_images(self.root)

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': data['file'].apply(lambda x: os.path.splitext(x)[0]),
            'path': data['path'] + os.path.sep + data['file'],
            'video': data['file'].str[7:10].astype(int),
            'id': data['path'].astype(int)
        })
        df = pd.merge(df, metadata, on='id', how='left')
        df.rename({'name': 'identity', 'sex': 'gender'}, axis=1, inplace=True)
        df = df.drop(['id', 'zoo', 'tracklets'], axis=1)
        return self.finalize_catalogue(df)



import pandas as pd
import matplotlib.pyplot as plt

def recognize_id_split(ids_train, ids_test):
    ids_train = set(ids_train)
    ids_test = set(ids_test)
    
    ids_test_only = ids_test - ids_train
    ids_joint = ids_train.intersection(ids_test)
    id_split = 'closed-set'
    if len(ids_joint) == 0:
        id_split = 'disjoint-set'
    elif len(ids_test_only) > 0:
        id_split = 'open-set'
    return id_split

def recognize_time_split(df_train, df_test, col_label='identity'):
    ids_train = set(df_train[col_label])
    ids_test = set(df_test[col_label])
    ids_joint = ids_train.intersection(ids_test)

    time_split = 'time-unaware'
    if 'date' in df_train.columns and 'date' in df_test.columns and sum(df_train['date'].isnull()) == 0 and sum(df_test['date'].isnull()) == 0:
        if max(df_train['date']) <= min(df_test['date']):
            time_split = 'time-cutoff'                
        else:
            change_split = True
            for identity in ids_joint:
                df_train_red = df_train[df_train[col_label] == identity]
                df_test_red = df_test[df_test[col_label] == identity]
                if len(df_train_red) > 0 and len(df_test_red) > 0 and max(df_train_red['date']) > min(df_test_red['date']):
                    change_split = False
                    break
            if change_split and len(ids_joint) > 0:
                time_split = 'time-proportion'
    return time_split

def extract_data_split(df, idx_train, idx_test, col_label='identity'):
    df_train = df.loc[idx_train]
    df_test = df.loc[idx_test]
    
    ids = set(df[col_label])
    ids_train = set(df_train[col_label])
    ids_test = set(df_test[col_label])
    ids_train_only = ids_train - ids_test
    ids_test_only = ids_test - ids_train
    
    id_split = recognize_id_split(ids_train, ids_test)
    time_split = recognize_time_split(df_train, df_test, col_label=col_label)

    return {
        'id_split': id_split,
        'time_split': time_split,
        'n': len(df),
        'n_train': len(df_train),
        'n_test': len(df_test),
        'n_train_only': sum([sum(df_train[col_label] == ids) for ids in ids_train_only]),
        'n_test_only': sum([sum(df_test[col_label] == ids) for ids in ids_test_only]),
        'n_unassigned': len(df)-len(df_train)-len(df_test),
        'n_ids': len(ids),
        'n_ids_train': len(ids_train),
        'n_ids_test': len(ids_test),
        'n_ids_joint': len(ids_train.intersection(ids_test)),
        'n_ids_train_only': len(ids_train_only),
        'n_ids_test_only': len(ids_test_only),
        'n_ids_unassigned': len(ids)-len(ids_train)-len(ids_test)+len(ids_train.intersection(ids_test)),
    }

def analyze_split(df, idx_train, idx_test, col_label='identity'):
    data = extract_data_split(df, idx_train, idx_test, col_label=col_label)
    print('Split: %s %s' % (data['time_split'], data['id_split']))
    print('Samples: train/test/unassigned/total = %d/%d/%d/%d' % (data['n_train'], data['n_test'], data['n_unassigned'], data['n']))
    print('Classes: train/test/unassigned/total = %d/%d/%d/%d' % (data['n_ids_train'], data['n_ids_test'], data['n_ids_unassigned'], data['n_ids']))
    print('Samples: train only/test only        = %d/%d' % (data['n_train_only'], data['n_test_only']))
    print('Classes: train only/test only/joint  = %d/%d/%d' % (data['n_ids_train_only'], data['n_ids_test_only'],  data['n_ids_joint']))
    print('')    
    print('Fraction of train set     = %1.2f%%' % (100*data['n_train']/data['n']))
    print('Fraction of test set only = %1.2f%%' % (100*data['n_test_only']/data['n']))

def visualize_split(df_train, df_test, selection=None, ylabels=True, col_label='identity'):
    if selection == None:
        selection = pd.concat((df_train, df_test))[col_label].sort_values().unique()
    elif isinstance(selection, int):
        selection = pd.concat((df_train, df_test))[col_label].sort_values().unique()[:selection]
    for i, identity in enumerate(selection):
        date_train = df_train[df_train[col_label] == identity]['date']
        plt.scatter(date_train, len(date_train)*[i], color='blue')
        date_test = df_test[df_test[col_label] == identity]['date']
        plt.scatter(date_test, len(date_test)*[i], color='black', marker='*')
    if ylabels:
        plt.yticks(range(len(selection)), selection)
        plt.ylabel('Turtle name')
    plt.scatter([], [], color='blue', label='Training set')
    plt.scatter([], [], color='black', marker='*', label='Testing set')
    plt.legend()


import os
import datetime
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class Cows2021(DatasetFactory):
    outdated_dataset = True
    summary = summary['Cows2021']
    url = 'https://data.bris.ac.uk/datasets/tar/4vnrca7qw1642qlwxjadp87h7.zip'
    archive = '4vnrca7qw1642qlwxjadp87h7.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
    
    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Extract information from the folder structure
        ii = (folders[2] == 'Identification') & (folders[3] == 'Test')
        folders = folders.loc[ii]
        data = data.loc[ii]

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': folders[4].astype(int),
        })
        df['date'] = df['path'].apply(lambda x: self.extract_date(x))
        return self.finalize_catalogue(df)

    def extract_date(self, x):
        x = os.path.split(x)[1]
        if x.startswith('image_'):
            x = x[6:]
        if x[7] == '_':
            x = x[8:]
        i1 = x.find('_')
        i2 = x[i1+1:].find('_')
        x = x[:i1+i2+1]
        return datetime.datetime.strptime(x, '%Y-%m-%d_%H-%M-%S').strftime('%Y-%m-%d %H:%M:%S')

class Cows2021v2(Cows2021):
    outdated_dataset = False

    def fix_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        # Replace the wrong identities and images
        replace_identity1 = [
            (164, 148),
            (105, 29)
        ]
        replace_identity2 = [
            ('image_0001226_2020-02-11_12-43-7_roi_001.jpg', 137, 107)
        ]
        df = self.fix_labels_replace_identity(df, replace_identity1)
        return self.fix_labels_replace_images(df, replace_identity2)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class StripeSpotter(DatasetFactory):
    summary = summary['StripeSpotter']

    @classmethod
    def _download(cls):
        downloads = [
            ('https://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/stripespotter/data-20110718.zip', 'data-20110718.zip'),
            ('https://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/stripespotter/data-20110718.z02', 'data-20110718.z02'),
            ('https://storage.googleapis.com/google-code-archive-downloads/v2/code.google.com/stripespotter/data-20110718.z01', 'data-20110718.z01'),
        ]
        for url, archive in downloads:
            utils.download_url(url, archive)

    @classmethod
    def _extract(cls):
        exception_text = '''Extracting works only on Linux. Please extract it manually.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#stripespotter'''
        if os.name == 'posix':
            os.system(f"zip -s- data-20110718.zip -O data-full.zip")
            if not os.path.exists('data-full.zip'):
                raise Exception('Download or extraction failed. Check if zip is installed.')
            os.system(f"unzip data-full.zip")
            os.remove('data-20110718.zip')
            os.remove('data-20110718.z01')
            os.remove('data-20110718.z02')
            os.remove('data-full.zip')
        else:
            raise Exception(exception_text)       

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)

        # Extract information about the images
        data['index'] = data['file'].str[-7:-4].astype(int)
        data = data[data['file'].str.startswith('img')]
        
        # Load additional information
        data_aux = pd.read_csv(os.path.join(self.root, 'data', 'SightingData.csv'))
        data = pd.merge(data, data_aux, how='left', left_on='index', right_on='#imgindex')
        data.loc[data['animal_name'].isnull(), 'animal_name'] = self.unknown_name
        
        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path':  data['path'] + os.path.sep + data['file'],
            'identity': data['animal_name'],
            'bbox': pd.Series([[int(a) for a in b.split(' ')] for b in data['roi']]),
            'orientation': data['flank'],
            'photo_quality': data['photo_quality'],
            'date': data['sighting_date']
        })
        return self.finalize_catalogue(df)  



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class CowDataset(DatasetFactory):
    summary = summary['CowDataset']
    url = 'https://figshare.com/ndownloader/files/31210192'
    archive = 'cow-dataset.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
        # Rename the folder with non-ASCII characters
        dirs = [x for x in os.listdir() if os.path.isdir(x)]
        if len(dirs) != 1:
            raise Exception('There should be only one directory after extracting the file.')
        os.rename(dirs[0], 'images')
    
    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        path = data['path'] + os.path.sep + data['file']
        date = [utils.get_image_date(os.path.join(self.root, p)) for p in path]

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path': path,
            'identity': folders[1].str.strip('cow_').astype(int),
            'date': date,
        })
        return self.finalize_catalogue(df)



import os
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class FriesianCattle2015(DatasetFactory):
    outdated_dataset = True
    summary = summary['FriesianCattle2015']
    url = 'https://data.bris.ac.uk/datasets/wurzq71kfm561ljahbwjhx9n3/wurzq71kfm561ljahbwjhx9n3.zip'
    archive = 'wurzq71kfm561ljahbwjhx9n3.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        
        # Extract information from the folder structure
        split = folders[1].replace({'Cows-testing': 'test', 'Cows-training': 'train'})
        identity = folders[2].str.strip('Cow').astype(int)

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(identity.astype(str) + split + data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': identity,
            'split': split
        })
        return self.finalize_catalogue(df)

class FriesianCattle2015v2(FriesianCattle2015):
    outdated_dataset = False

    def fix_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        # Remove all identities in training as they are duplicates
        idx_remove = ['Cows-training' in path for path in df.path]
        df = df[~np.array(idx_remove)]

        # Remove specified individuals as they are duplicates
        identities_to_remove = [19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 31, 32, 33, 37]
        return self.fix_labels_remove_identity(df, identities_to_remove)


class FriesianCattle2017(DatasetFactory):
    summary = summary['FriesianCattle2017']
    url = 'https://data.bris.ac.uk/datasets/2yizcfbkuv4352pzc32n54371r/2yizcfbkuv4352pzc32n54371r.zip'
    archive = '2yizcfbkuv4352pzc32n54371r.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': folders[1].astype(int),
        })
        return self.finalize_catalogue(df)



import os
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class ELPephants(DatasetFactory):
    summary = summary['ELPephants']
    archive = 'ELPephant (elephant ID system).zip'

    @classmethod
    def _download(cls):
        exception_text = '''Dataset must be downloaded manually.
            Check https://wildlifedatasets.github.io/wildlife-datasets/preprocessing#elpephants'''
        raise Exception(exception_text)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)

        # Create the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': data['file'].apply(self.extract_identity),
            'date': data['file'].apply(self.extract_date),
            'orientation': data['file'].apply(self.extract_orientation),
        })

        # Add training and testing split
        path_txt = utils.find_images(self.root, img_extensions='.txt')
        idx_train = np.where(path_txt['file'] == 'train.txt')[0]
        idx_test = np.where(path_txt['file'] == 'val.txt')[0]
        if len(idx_train) == 1 and len(idx_test) == 1:            
            data_train = pd.read_csv(os.path.join(self.root, path_txt['path'].iloc[idx_train[0]], 'train.txt'), header=None, sep='\t')
            data_train = data_train[1].to_numpy()
            data_test = pd.read_csv(os.path.join(self.root, path_txt['path'].iloc[idx_test[0]], 'val.txt'), header=None, sep='\t')
            data_test = data_test[1].to_numpy()
            df['original_split'] = data['file'].apply(lambda x: utils.get_split(x, data_train, data_test))

        # Finalize the dataframe
        return self.finalize_catalogue(df)

    def extract_identity(self, x):
        return int(x.strip().split('_')[0])

    def extract_date(self, x):
        for y in x.lower().split('_')[::-1]:
            date = self.extract_date_part(y)
            if isinstance(date, str):
                return date
        return np.nan
        
    def extract_date_part(self, x):
        i_end = -np.inf
        conversion = {
            'january': 1, 'february': 2, 'march': 3, 'april': 4,
            'may': 5, 'june': 6, 'july': 7, 'august': 8,
            'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4,
            'may': 5, 'jun': 6, 'jul': 7, 'aug': 8,
            'sep': 9, 'oct': 10, 'nov': 11, 'dec': 12
            }
        # Extract month
        for month_try in conversion:
            if month_try in x:
                i = x.index(month_try)
                if i > i_end:
                    month = month_try
                    i_end = i
        # Month not found
        if i_end == -np.inf:
            return np.nan
        # Extract string after month and remove extension
        year = x[i_end+len(month):]
        year = os.path.splitext(year)[0]
        year = year.strip()
        # Extract year from the beginning or end
        if len(year) >= 4:
            if represents_int(year[:4]):
                return f'{int(year[:4])}-{conversion[month]}-1'
            elif represents_int(year[-4:]):
                return f'{int(year[-4:])}-{conversion[month]}-1'
        return np.nan
        
    def extract_orientation(self, x):
        x = x.lower()
        for orientation in ['left', 'right', 'front']:
            if orientation in x:
                return orientation
        return np.nan         

def represents_int(s):
    try: 
        int(s)
    except ValueError:
        return False
    else:
        return True



import numpy as np
import sklearn.metrics as skm
import pandas as pd
from typing import List, Tuple, Union, Optional

def unify_types(
        y_true: List,
        y_pred: List,
        new_class: Optional[Union[int, str]] = None,
    ) -> Tuple[List, List, Union[int, str]]:
    """Unifies label types.

    If `new_class` is string and labels integers (or the other way round),
    it converts them to integers to allow to us numpy arrays.
    If it is `None`, it only checks for compatibility of `y_true` and `y_pred`.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Optional[Union[int, str]]): Name of the new class.

    Returns:
        Unified label types.
    """

    if new_class is not None:
        y_all = list(set(y_true).union(set(y_pred)) - set([new_class]))
    else:
        y_all = list(set(y_true).union(set(y_pred)))
    is_integer = pd.api.types.is_integer_dtype(pd.Series(y_all))
    is_string = pd.api.types.is_string_dtype(pd.Series(y_all))
    if not is_integer and not is_string:
        raise(Exception('Labels have mixed types. Convert all to int or str.'))
    if is_string and isinstance(new_class, int):
        encoder = {new_class: new_class}
        for i, y in enumerate(y_all):
            encoder[y] = new_class + 1 + i
        y_true = [encoder[y] for y in y_true]
        y_pred = [encoder[y] for y in y_pred]
    if is_integer and isinstance(new_class, str):
        encoder = {new_class: np.min(y_all) - 1}
        for i, y in enumerate(y_all):
            encoder[y] = y
        y_true = [encoder[y] for y in y_true]
        y_pred = [encoder[y] for y in y_pred]
        new_class = encoder[new_class]
    return y_true, y_pred, new_class


def accuracy(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str] = None,
    ) -> float:
    """Computes the accuracy.

    If `new_class` is specified, it handles it as a additional class.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed accuracy.
    """

    y_true, y_pred, new_class = unify_types(y_true, y_pred, new_class)
    return np.mean(np.array(y_pred) == np.array(y_true))

def balanced_accuracy(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str] = None,
    ) -> float:
    """Computes the balanced accuracy.

    If `new_class` is specified, it handles it as a additional class.
    Each class has the same weight irregardless of the number of samples.
    If equals to macro recall unless there are predictions
    contain classes not in the true labels.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed balanced accuracy.
    """

    y_true, y_pred, new_class = unify_types(y_true, y_pred, new_class)    
    C = skm.confusion_matrix(y_true, y_pred)
    with np.errstate(divide="ignore", invalid="ignore"):
        per_class = np.diag(C) / C.sum(axis=1)
    return np.mean(per_class[~np.isnan(per_class)])

def class_average_accuracy(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str] = None,
    ) -> float:
    """Computes the class average accuracy.

    If `new_class` is specified, it handles it as a additional class.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed class average accuracy.
    """
    
    y_true, y_pred, new_class = unify_types(y_true, y_pred, new_class)
    C = skm.multilabel_confusion_matrix(y_true, y_pred)
    return np.mean([C_i[0,0]+C_i[1,1] for C_i in C]) / np.sum(C[0])

def precision(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str] = None,
    ) -> float:
    """Computes the (macro-averaged) precision.

    If `new_class` is specified, it handles it as a additional class.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed precision.
    """

    y_true, y_pred, new_class = unify_types(y_true, y_pred, new_class)
    return skm.precision_score(y_true, y_pred, average='macro')

def recall(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str] = None,
        ignore_empty: bool = False
    ) -> float:
    """Computes the (macro-averaged) recall.

    If `new_class` is specified, it handles it as a additional class.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.
        ignore_empty (bool): Whether classes not in true labels should be ignored for averaging.

    Returns:
        Computed recall.
    """

    y_true, y_pred, new_class = unify_types(y_true, y_pred, new_class)
    if ignore_empty:
        C = skm.multilabel_confusion_matrix(y_true, y_pred)
        return np.mean([C_i[1,1]/(C_i[1,0]+C_i[1,1]) for C_i in C if C_i[1,0]+C_i[1,1] > 0])
    else:
        return skm.recall_score(y_true, y_pred, average='macro')

def f1(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str] = None,
    ) -> float:
    """Computes the (macro-averaged) F1 score.

    If `new_class` is specified, it handles it as a additional class.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed F1 score.
    """

    y_true, y_pred, new_class = unify_types(y_true, y_pred, new_class)
    return skm.f1_score(y_true, y_pred, average='macro')

def accuracy_known_samples(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str],
    ) -> float:
    """Computes the accuracy on known samples.

    Samples of new (unknown) class are ignored.
    Returns `np.nan` if there are no known samples.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed accuracy on known samples.
    """

    y_true, y_pred, new_class = unify_types(y_true, y_pred, new_class)
    y_true = np.array(y_true)    
    y_pred = np.array(y_pred)
    known = y_true != new_class
    if sum(known) > 0:
        return np.mean(y_true[known] == y_pred[known])
    else:
        return np.nan

def accuracy_unknown_samples(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str],
    ) -> float:
    """Computes the accuracy on new (unknown) samples.

    Samples of known classes are ignored.
    Returns `np.nan` if there are no unknown samples.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed accuracy on unknown samples.
    """

    y_true, y_pred, new_class = unify_types(y_true, y_pred, new_class)
    y_true = np.array(y_true)    
    y_pred = np.array(y_pred)
    unknown = y_true == new_class
    if sum(unknown) > 0:
        return np.mean(y_true[unknown] == y_pred[unknown])
    else:
        return np.nan
    
def normalized_accuracy(
        y_true: List,
        y_pred: List,
        new_class: Union[int, str],
        mu: float
    ) -> float:
    """Computes the normalized accuracy.

    It is mu*accuracy on known samples + (1-mu)*accuracy on unknown samples.
    Returns `np.nan` if there are no known samples or no unknown samples.

    Args:
        y_true (List): List of true labels.
        y_pred (List): List of predictions.
        new_class (Union[int, str]): Name of the new class.
        mu (float): Coefficient of the convex combination.

    Returns:
        Computed normalized accuracy.
    """

    aks = accuracy_known_samples(y_true, y_pred, new_class)
    aus = accuracy_unknown_samples(y_true, y_pred, new_class)
    return mu*aks + (1-mu)*aus

def average_precision(
        y_true: Union[int, str],
        y_pred: List
    ) -> float:
    """Computes the average precision (for one sample).

    It computes AP@K, where K is the length of `y_pred`.

    Args:
        y_true (Union[int, str]): The true label.
        y_pred (List): Ranked predictions. First elements have a better rank.

    Returns:
        Computed average precision.
    """

    unify_types([y_true], y_pred)
    a = np.array(y_pred) == y_true
    b = np.linspace(1, 0, len(y_pred))
    if sum(a) == 0:
        return 0
    else:
        return skm.average_precision_score(a, b)

def mean_average_precision(
        y_true: List,
        y_pred: List[List]
    ):
    """Computes the mean average precision (for multiple samples).

    It computes MAP@K, where K is the length of `y_pred[0]`.

    Args:
        y_true (List): List of true labels.
        y_pred (List[List]): List of ranked predictions. First elements have a better rank.

    Returns:
        Computed mean average precision.
    """

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    return np.mean([average_precision(y_t, y_p) for y_t, y_p in zip(y_true, y_pred)])

def auc_roc_new_class(
        y_true: List,
        y_score: List,
        new_class: Union[int, str],
    ) -> float:
    """Computes the area under ROC curve for detecting new individuals.

    High score suggests known individual.
    Low score suggests new (unknown) individual.
    It converts the problem to binary (is it new individual or not)
    and computes AUC for it.

    Args:
        y_true (List): List of true labels.
        y_score (List): List of scores.
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed arena under ROC curve.
    """

    y_true, _, new_class = unify_types(y_true, [], new_class)    
    a = np.array(y_true) == new_class
    b = -np.array(y_score)
    return skm.roc_auc_score(a, b)

def BAKS(
        y_true: List,
        y_pred: List,
        identity_test_only: List,
    ) -> float:
    """Computes BAKS (balanced accuracy on known samples).

    It ignores `identity_test_only` because they are unknown identities.

    Args:
        y_true (List): List of true labels.
        y_score (List): List of scores.
        identity_test_only (List): List of new identities (only in the testing set).

    Returns:
        Computed BAKS.
    """

    # Need to keep the object type due to mixed arrays
    y_true = np.array(y_true, dtype=object)
    y_pred = np.array(y_pred, dtype=object)
    identity_test_only = np.array(identity_test_only, dtype=object)

    # Remove data in identity_test_only
    idx = np.where(~np.isin(y_true, identity_test_only))[0]
    y_true_idx = y_true[idx]
    y_pred_idx = y_pred[idx]
    if len(y_true_idx) == 0:
        return np.nan
    
    # Check if the remaining types are compatible
    unify_types(y_true_idx, y_pred_idx)
    df = pd.DataFrame({'y_true': y_true_idx, 'y_pred': y_pred_idx})

    # Compute the balanced accuracy
    accuracy = 0
    for _, df_identity in df.groupby('y_true'):
        accuracy += 1 / df['y_true'].nunique() * np.mean(df_identity['y_pred'] == df_identity['y_true'])
    return accuracy

def BAUS(
        y_true: List,
        y_pred: List,
        identity_test_only: List,
        new_class: Union[int, str]
    ) -> float:
    """Computes BAUS (balanced accuracy on unknown samples).

    It handles only `identity_test_only` because they are unknown identities.

    Args:
        y_true (List): List of true labels.
        y_score (List): List of scores.
        identity_test_only (List): List of new identities (only in the testing set).
        new_class (Union[int, str]): Name of the new class.

    Returns:
        Computed BAUS.
    """

    # Need to keep the object type due to mixed arrays
    y_true = np.array(y_true, dtype=object)
    y_pred = np.array(y_pred, dtype=object)
    identity_test_only = np.array(identity_test_only, dtype=object)

    # Remove data not in identity_test_only
    idx = np.where(np.isin(y_true, identity_test_only))[0]
    y_true_idx = y_true[idx]
    y_pred_idx = y_pred[idx]
    if len(y_true_idx) == 0:
        return np.nan

    # Check if the remaining types are compatible
    y_true_idx, y_pred_idx, new_class = unify_types(y_true_idx, y_pred_idx, new_class)
    df = pd.DataFrame({'y_true': y_true_idx, 'y_pred': y_pred_idx})

    # Compute the balanced accuracy
    accuracy = 0
    for _, df_identity in df.groupby('y_true'):
        accuracy += 1 / df['y_true'].nunique() * np.mean(df_identity['y_pred'] == new_class)
    return accuracy



import os
import json
import string
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class IPanda50(DatasetFactory):
    summary = summary['IPanda50']
    downloads = [
        ('https://drive.google.com/uc?id=1nkh-g6a8JvWy-XsMaZqrN2AXoPlaXuFg', 'iPanda50-images.zip'),
        ('https://drive.google.com/uc?id=1gVREtFWkNec4xwqOyKkpuIQIyWU_Y_Ob', 'iPanda50-split.zip'),
        ('https://drive.google.com/uc?id=1jdACN98uOxedZDT-6X3rpbooLAAUEbNY', 'iPanda50-eyes-labels.zip'),
    ]

    @classmethod
    def _download(cls):
        exception_text = '''Download failed. GDown quota probably reached. Download dataset manually.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#ipanda50'''
        for url, archive in cls.downloads:
            utils.gdown_download(url, archive, exception_text=exception_text)

    @classmethod
    def _extract(cls):
        for url, archive in cls.downloads:
            utils.extract_archive(archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Extract keypoint information about eyes
        keypoints = []
        for path in data['path'] + os.path.sep + data['file']:
            path_split = os.path.normpath(path).split(os.path.sep)
            path_json = os.path.join('iPanda50-eyes-labels', path_split[1], os.path.splitext(path_split[2])[0] + '.json')
            keypoints_part = np.full(8, np.nan)        
            if os.path.exists(os.path.join(self.root, path_json)):
                with open(os.path.join(self.root, path_json)) as file:
                    keypoints_file = json.load(file)['shapes']
                    if keypoints_file[0]['label'] == 'right_eye':
                        keypoints_part[0:4] = np.reshape(keypoints_file[0]['points'], 4)
                    if keypoints_file[0]['label'] == 'left_eye':
                        keypoints_part[4:8] = np.reshape(keypoints_file[0]['points'], 4)
                    if len(keypoints_file) == 2 and keypoints_file[1]['label'] == 'right_eye':
                        keypoints_part[0:4] = np.reshape(keypoints_file[1]['points'], 4)
                    if len(keypoints_file) == 2 and keypoints_file[1]['label'] == 'left_eye':
                        keypoints_part[4:8] = np.reshape(keypoints_file[1]['points'], 4)
            keypoints.append(list(keypoints_part))
        
        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': folders[1],
            'keypoints': keypoints
            })

        # Remove non-ASCII characters while keeping backwards compatibility
        file_name = os.path.join(self.root, 'changes.csv')
        if os.path.exists(file_name):
            # Files were already renamed, change image_id to keep backward compability
            df_changes = pd.read_csv(file_name)
            id_changes = {}
            for _, df_row in df_changes.iterrows():
                id_changes[df_row['id_new']] = df_row['id_old']
            df.replace(id_changes, inplace=True)
        else:
            # Rename files, keep original image_id, create list of changes
            ids_old = []
            ids_new = []
            for _, df_row in df.iterrows():
                path_new = ''.join(c for c in df_row['path'] if c in string.printable)
                # Check if there are non-ASCII characters
                if path_new != df_row['path']:
                    # Rename files and df
                    os.rename(os.path.join(self.root, df_row['path']), os.path.join(self.root, path_new))
                    df_row['path'] = path_new
                    # Save changes in image_id
                    ids_old.append(df_row['image_id'])
                    ids_new.append(utils.create_id(pd.Series(os.path.split(df_row['path'])[-1])).iloc[0])
            if len(df) != df['path'].nunique():
                raise(Exception("Non-unique names. Something went wrong when renaming."))
            pd.DataFrame({'id_old': ids_old, 'id_new': ids_new}).to_csv(file_name)

        # Finalize the dataframe        
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class BristolGorillas2020(DatasetFactory):
    summary = summary['BristolGorillas2020']
    url = 'https://data.bris.ac.uk/datasets/tar/jf0859kboy8k2ufv60dqeb2t8.zip'
    archive = 'jf0859kboy8k2ufv60dqeb2t8.zip'

    @classmethod
    def _download(cls):
        command = f"wget -c -q {cls.url}"
        exception_text = '''Download works only on Linux. Please download it manually.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#bristolgorillas2020'''
        if os.name == 'posix':
            os.system(command)
        else:
            raise Exception(exception_text)

    @classmethod
    def _extract(cls):
        exception_text = '''Extracting resulted in the expection below.
            This may have happened that because the file was not downloaded correctly.
            Download manually or try to call (the download should resume):
            datasets.BristolGorillas2020.get_data(root, force=True)'''
        try:
            utils.extract_archive(cls.archive, delete=True)
        except Exception as e:
            print(e)            
            raise Exception(exception_text)
    
    def create_catalogue(self) -> pd.DataFrame:
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        n_folders = max(folders.columns)

        # Restrict to correct images
        idx = folders[n_folders-2] != 'videos'
        data = data[idx]
        folders = folders[idx]

        # Create the dataframe
        df1 = pd.DataFrame({
            'path': data['path'] + os.path.sep + data['file'],
            'original_split': folders[n_folders-2],
        })
        
        # Add bounding boxes (either load them or create and save them)
        identity_conversion = {
            0: 'afia',
            1: 'ayana',
            2: 'jock',
            3: 'kala',
            4: 'kera',
            5: 'kukuena',
            6: 'touni'
        }

        identity_all = []
        bbox_all = []
        path_all = []
        for i in range(len(df1)):
            path_img = os.path.join(self.root, df1['path'].iloc[i])
            path_bbox = os.path.splitext(path_img)[0] + '.txt'
            path_size = os.path.splitext(path_img)[0] + '_size.txt'

            # Load image size
            if os.path.exists(path_size):
                with open(path_size, 'r') as file:
                    line = file.readline()
                w, h = [int(num) for num in line.split()]
            else:
                img = utils.load_image(path_img)
                w, h = img.size
                with open(path_size, 'w') as file:
                    file.write(f'{w} {h}\n')

            # Load bounding boxes
            with open(path_bbox, 'r') as file:
                lines = [line.rstrip() for line in file]
            for line in lines:
                if len(line) > 0:
                    identity = identity_conversion[int(line[0])]
                    bbox = [float(num) for num in line.split()[1:]]
                    bbox = utils.yolo_to_pascalvoc(*bbox, w, h)
                    bbox = [bbox[0], bbox[1], bbox[2]-bbox[0], bbox[3]-bbox[1]]
                    identity_all.append(identity)
                    bbox_all.append(bbox)
                    path_all.append(df1['path'].iloc[i])

        # Merge bounding boxes into paths and remove images with no bounding boxes
        df2 = pd.DataFrame({'identity': identity_all, 'bbox': bbox_all, 'path': path_all})
        df = pd.merge(df1, df2, on='path', how='left')
        df = df[~df['identity'].isnull()]

        # Finalize catalogue
        df['image_id'] = utils.create_id(df['path'].apply(lambda x: os.path.basename(x)) + df['identity'])
        return self.finalize_catalogue(df)



import pandas as pd
from . import utils
from .datasets_wildme import DatasetFactoryWildMe
from .summary import summary

class WhaleSharkID(DatasetFactoryWildMe):
    summary = summary['WhaleSharkID']
    url = 'https://lilawildlife.blob.core.windows.net/lila-wildlife/wild-me/whaleshark.coco.tar.gz'
    archive = 'whaleshark.coco.tar.gz'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        return self.create_catalogue_wildme('whaleshark', 2020)



import os
import json
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary


class CzechLynx(DatasetFactory):
    summary = summary['CzechLynx']
    archive = 'czechlynx.zip'

    @classmethod
    def _download(cls):
        command = f"datasets download -d picekl/czechlynx --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#czechlynx'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self, split: str = 'split-geo_aware') -> pd.DataFrame:
        """
        Creates a comprehensive catalogue DataFrame for the CzechLynx dataset.

        This method reads the dataset's metadata file, processes essential fields required
        by the framework, and returns a detailed dataframe containing information about
        each animal observation and its associated metadata.

        Args:
            split (str): One of the following options:
                - 'split-geo_aware': Split based on spatial location.
                - 'split-time_open': Time-aware split (open-world).
                - 'split-time_closed': Time-aware split (closed-world).
                Default is 'split-geo_aware'.

        Returns:
            pd.DataFrame: A dataframe including the following columns:

                - image_id (str): Unique identifier for each row.
                - identity (str): Identity of the depicted individual animal.
                                  If unknown, standardized to the framework’s unknown label.
                - path (str): Relative path to the image file.
                - source (str): Dataset partition or region (e.g., 'beskydy').
                - date (datetime): Observation date. Converted to `datetime`.
                - relative_age (float): Age estimate relative to first appearance.
                - encounter (int): Unique encounter identifier.
                - coat_pattern (str): Coat pattern or marking description.
                - location (str): Specific trap or site location.
                - cell_code (str): Grid cell reference (e.g., 10km spatial cell).
                - latitude (float): Latitude of the observation point.
                - longitude (float): Longitude of the observation point.
                - trap_id (str): Unique identifier for the camera trap.
                - split-geo_aware (str): Spatial split category.
                - split-time_open (str): Time-aware open-world split.
                - split-time_closed (str): Time-aware closed-world split.

        Notes:
            - The `unique_name` column is renamed to `identity` and then dropped.
            - No filtering is applied to the split; the selected column is kept for downstream use.
            - All metadata columns are retained except explicitly replaced ones.
        """

        # Load metadata
        metadata_path = os.path.join(self.root, 'metadata.csv')
        df = pd.read_csv(metadata_path)

        # Add required columns
        df['image_id'] = df.index.astype(str)
        df['identity'] = df['unique_name']
        idx = ~df['date'].isnull()
        df.loc[idx, 'date'] = df.loc[idx, 'date'].apply(lambda x: str(x)[6:10] + '-' + str(x)[3:5] + '-' + str(x)[:2])
        df.drop(columns=['unique_name'], inplace=True)

        # Keep only selected split column, rename it
        df['original_split'] = df[split]
        df.drop(columns=['unique_name', 'split-geo_aware', 'split-time_open', 'split-time_closed'],
                errors='ignore', inplace=True)

        return self.finalize_catalogue(df)





import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class MPDD(DatasetFactory):
    summary = summary['MPDD']
    url = 'https://prod-dcd-datasets-cache-zipfiles.s3.eu-west-1.amazonaws.com/v5j6m8dzhv-1.zip'
    archive = 'MPDD.zip'
    
    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
        utils.extract_archive(os.path.join('Multi-pose dog dataset', 'MPDD.zip'), delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        identity = data['file'].apply(lambda x: int(x.split('_')[0]))
        
        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': data['file'],
            'path': data['path'] + os.path.sep + data['file'],
            'identity': identity,
            'original_split': folders[2]
        })
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class WildRaptorID(DatasetFactory):
    summary = summary['WildRaptorID']
    url = 'https://zenodo.org/records/16731160/files/wild_raptor_id.zip?download=1'
    archive = 'wild_raptor_id.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        n_folders = max(folders.columns)

        data['file_size'] = [os.path.getsize(os.path.join(self.root, x['path'], x['file'])) for _, x in data.iterrows()]
        data['identity'] = folders[n_folders].apply(lambda x: x.split('_')[0])

        # Very rough way of checking that the files with the same cols are duplicates
        cols = ['file', 'identity']
        for _, df_red in data.groupby(cols):
            if len(df_red) > 1:
                if df_red['file_size'].nunique() != 1:
                    raise Exception('The expected duplicates are not duplicates')
                
        # Remove the duplicates
        data = data.drop_duplicates(subset=cols)

        # Add date
        file = data['file'].apply(lambda x: os.path.splitext(x)[0][1:-1])
        date = file.apply(lambda x: x.split('_')[0])
        date = date.apply(lambda x: f'{x[:4]}-{x[4:6]}-{x[6:8]} {x[8:10]}:{x[10:12]}')
        data['date'] = date

        # Add video        
        for i, (_, df_group) in enumerate(data.groupby(['identity', 'date'])):
            data.loc[df_group.index, 'video'] = i
        data['video'] = data['video'].astype(int)
        
        # Finalize the dataframe
        data['image_id'] = data['identity'] + '_' + file
        data['path'] = data['path'] + os.path.sep + data['file']
        data = data.drop(['file', 'file_size'], axis=1)
        return self.finalize_catalogue(data)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class OpenCows2020(DatasetFactory):
    summary = summary['OpenCows2020']
    url = 'https://data.bris.ac.uk/datasets/tar/10m32xl88x2b61zlkkgz3fml17.zip'
    archive = '10m32xl88x2b61zlkkgz3fml17.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Select only re-identification dataset
        reid = folders[1] == 'identification'
        folders, data = folders[reid], data[reid]

        # Extract information from the folder structure
        split = folders[3]
        identity = folders[4]

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(identity.astype(str) + split + data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': identity,
            'original_split': split
        })
        return self.finalize_catalogue(df)    



import os
import shutil
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class NOAARightWhale(DatasetFactory):
    summary = summary['NOAARightWhale']
    archive = 'noaa-right-whale-recognition.zip'

    @classmethod
    def _download(cls):
        command = f"competitions download -c noaa-right-whale-recognition --force"
        exception_text = '''Kaggle terms must be agreed with.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#noaarightwhale'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        try:
            utils.extract_archive(cls.archive, delete=True)
            utils.extract_archive('imgs.zip', delete=True)
            # Move misplaced image
            shutil.move('w_7489.jpg', 'imgs')
            os.remove('w_7489.jpg.zip')
        except:
            exception_text = '''Extracting failed.
                Either the download was not completed or the Kaggle terms were not agreed with.
                Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#noaarightwhale'''
            raise Exception(exception_text)

    def create_catalogue(self) -> pd.DataFrame:
        # Load information about the training dataset
        data = pd.read_csv(os.path.join(self.root, 'train.csv'))
        df1 = pd.DataFrame({
            'image_id': data['Image'].str.split('.', expand=True)[0].str.strip('w_').astype(int),
            'path': 'imgs' + os.path.sep + data['Image'],
            'identity': data['whaleID'],
            'original_split': 'train'
            })

        # Load information about the testing dataset
        data = pd.read_csv(os.path.join(self.root, 'sample_submission.csv'))
        df2 = pd.DataFrame({
            'image_id': data['Image'].str.split('.', expand=True)[0].str.strip('w_').astype(int),
            'path': 'imgs' + os.path.sep + data['Image'],
            'identity': self.unknown_name,
            'original_split': 'test'
            })
        
        # Finalize the dataframe
        df = pd.concat([df1, df2])
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class SouthernProvinceTurtles(DatasetFactory):
    archive = 'southernprovinceturtles.zip'
    summary = summary['SouthernProvinceTurtles']

    @classmethod
    def _download(cls):
        command = f"datasets download -d wildlifedatasets/southernprovinceturtles --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#southernprovinceturtles'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        file_name = os.path.join(self.root, 'annotations.csv')
        data = pd.read_csv(file_name)

        df = pd.DataFrame({
            'image_id': range(len(data)),
            'path': 'data' + os.path.sep + data['image_name'],
            'identity': data['identity'],
            'orientation': data['orientation'],
            'daytime': data['daytime']
        })
        if 'bbox_x' in data.columns:
            df['bbox'] = data[['bbox_x', 'bbox_y', 'bbox_width', 'bbox_height']].values.tolist()
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class DogFaceNet(DatasetFactory):
    summary = summary['DogFaceNet']
    url = 'https://github.com/GuillaumeMougeot/DogFaceNet/releases/download/dataset/DogFaceNet_Dataset_224_1.zip'
    archive = 'DogFaceNet_Dataset_224_1.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
    
    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': folders[1].astype(int),
        })
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets_wildme import DatasetFactoryWildMe
from .summary import summary

class BelugaID(DatasetFactoryWildMe):
    outdated_dataset = True
    summary = summary['BelugaID']
    downloads = [
        ('https://lilawildlife.blob.core.windows.net/lila-wildlife/wild-me/beluga.coco.tar.gz', 'beluga.coco.tar.gz'),
        ('https://lilawildlife.blob.core.windows.net/lila-wildlife/wild-me/beluga-id-test.zip', 'beluga-id-test.zip'),
    ]

    @classmethod
    def _download(cls):
        for url, archive in cls.downloads:
            utils.download_url(url, archive)

    @classmethod
    def _extract(cls):
        for url, archive in cls.downloads:
            archive_name = archive.split('.')[0]
            utils.extract_archive(archive, archive_name, delete=True)
    
    def create_catalogue(self) -> pd.DataFrame:
        return self.create_catalogue_wildme(os.path.join('beluga', 'beluga'), 2022)


class BelugaIDv2(BelugaID):
    outdated_dataset = False
    
    def create_catalogue(self) -> pd.DataFrame:
        # Use the original data
        df_train = self.create_catalogue_wildme(os.path.join('beluga', 'beluga'), 2022)
        df_train['original_split'] = 'train'

        # Get the conversion of identities between original and new data
        data_train = pd.read_csv(os.path.join(self.root, 'beluga-id-test', 'private_train_metadata.csv'))
        id_conversion = {}
        for old_id, data_train_red in data_train.groupby('original_whale_id'):
            if data_train_red['whale_id'].nunique() != 1:
                raise Exception('Conversion of old to new whale_id is not unique.')
            id_conversion[old_id] = data_train_red['whale_id'].iloc[0]

        # Add the new data
        data_test = pd.read_csv(os.path.join(self.root, 'beluga-id-test', 'private_test_metadata.csv'))
        df_test = pd.DataFrame({
            'path': data_test['path'].apply(lambda x: os.path.join('beluga-id-test', 'code-execution', 'images', os.path.split(x)[-1])),
            'image_id': range(len(data_train), len(data_train)+len(data_test)),
            'identity': data_test['original_whale_id'].apply(lambda x: id_conversion.get(x, 'unknown')),
            'date': data_test['timestamp'],
            'species': 'beluga_whale',
            'orientation': data_test['viewpoint'].replace({'top': 'up'}),
            'original_split': 'test'
        })
        
        # Finalize the dataframe
        df = pd.concat((df_train, df_test)).reset_index(drop=True)
        return self.finalize_catalogue(df)



import os
import shutil
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class CTai(DatasetFactory):
    summary = summary['CTai']
    url = 'https://github.com/cvjena/chimpanzee_faces/archive/refs/heads/master.zip'
    archive = 'master.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
        shutil.rmtree('chimpanzee_faces-master/datasets_cropped_chimpanzee_faces/data_CZoo')

    def create_catalogue(self) -> pd.DataFrame:            
        # Load information about the dataset
        path = os.path.join('chimpanzee_faces-master', 'datasets_cropped_chimpanzee_faces', 'data_CTai',)
        data = pd.read_csv(os.path.join(self.root, path, 'annotations_ctai.txt'), header=None, sep=' ')
        
        # Extract keypoints from the information
        keypoints = data[[11, 12, 14, 15, 17, 18, 20, 21, 23, 24]].to_numpy()
        keypoints[np.isinf(keypoints)] = np.nan
        keypoints = pd.Series(list(keypoints))
        
        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': pd.Series(range(len(data))),
            'path': path + os.path.sep + data[1],
            'identity': data[3],
            'keypoints': keypoints,
            'age': data[5],
            'age_group': data[7],
            'gender': data[9],
        })
        return self.finalize_catalogue(df)

    def fix_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        # Replace the wrong identities
        replace_identity = [
            ('Adult', self.unknown_name),
            ('Akouba', 'Akrouba'),
            ('Freddy', 'Fredy'),
            ('Ibrahiim', 'Ibrahim'),
            ('Liliou', 'Lilou'),
            ('Wapii', 'Wapi'),
            ('Woodstiock', 'Woodstock')
        ]
        return self.fix_labels_replace_identity(df, replace_identity)



import os
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class ZindiTurtleRecall(DatasetFactory):
    summary = summary['ZindiTurtleRecall']

    @classmethod
    def _download(cls):
        downloads = [
            ('https://storage.googleapis.com/dm-turtle-recall/train.csv', 'train.csv'),
            ('https://storage.googleapis.com/dm-turtle-recall/extra_images.csv', 'extra_images.csv'),
            ('https://storage.googleapis.com/dm-turtle-recall/test.csv', 'test.csv'),
            ('https://storage.googleapis.com/dm-turtle-recall/images.tar', 'images.tar'),
        ]
        for url, file in downloads:
            utils.download_url(url, file)

    @classmethod
    def _extract(cls):
        utils.extract_archive('images.tar', 'images', delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Load information about the training images
        data_train = pd.read_csv(os.path.join(self.root, 'train.csv'))
        data_train['split'] = 'train'

        # Load information about the testing images
        data_test = pd.read_csv(os.path.join(self.root, 'test.csv'))
        data_test['split'] = 'test'

        # Load information about the additional images
        data_extra = pd.read_csv(os.path.join(self.root, 'extra_images.csv'))
        data_extra['split'] = np.nan        

        # Finalize the dataframe
        data = pd.concat([data_train, data_test, data_extra])
        data = data.reset_index(drop=True)        
        data.loc[data['turtle_id'].isnull(), 'turtle_id'] = self.unknown_name
        df = pd.DataFrame({
            'image_id': data['image_id'],
            'path': 'images' + os.path.sep + data['image_id'] + '.JPG',
            'identity': data['turtle_id'],
            'orientation': data['image_location'].str.lower(),
            'original_split': data['split'],
        })
        return self.finalize_catalogue(df)


import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class SeaStarReID2023(DatasetFactory):
    summary = summary['SeaStarReID2023']
    url = 'https://storage.googleapis.com/public-datasets-lila/sea-star-re-id/sea-star-re-id.zip'
    archive = 'sea-star-re-id.zip'
    
    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        species = folders[1].str[:4].replace({'Anau': 'Anthenea australiae', 'Asru': 'Asteria rubens'})
        path = data['path'] + os.path.sep + data['file']
        date = [utils.get_image_date(os.path.join(self.root, p)) for p in path]

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': utils.create_id(data['file']),
            'path': path,
            'identity': folders[1],
            'species': species,
            'date': date
        })
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class AAUZebraFish(DatasetFactory):
    summary = summary['AAUZebraFish']
    archive = 'aau-zebrafish-reid.zip'

    @classmethod
    def _download(cls):
        command = f"datasets download -d aalborguniversity/aau-zebrafish-reid"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#aauzebrafish'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
        
    def create_catalogue(self) -> pd.DataFrame:
        data = pd.read_csv(os.path.join(self.root, 'annotations.csv'), sep=';')

        # Modify the bounding boxes into the required format
        columns_bbox = [
            'Upper left corner X',
            'Upper left corner Y',
            'Lower right corner X',
            'Lower right corner Y',
        ]
        bbox = data[columns_bbox].to_numpy()
        bbox[:,2] = bbox[:,2] - bbox[:,0]
        bbox[:,3] = bbox[:,3] - bbox[:,1]
        bbox = pd.Series(list(bbox))

        # Split additional data into a separate structure
        attributes = data['Right,Turning,Occlusion,Glitch'].str.split(',', expand=True)
        attributes.drop([0], axis=1, inplace=True)
        attributes.columns = ['turning', 'occlusion', 'glitch']
        attributes = attributes.astype(int).astype(bool)

        # Split additional data into a separate structure
        orientation = data['Right,Turning,Occlusion,Glitch'].str.split(',', expand=True)[0]
        orientation.replace('1', 'right', inplace=True)
        orientation.replace('0', 'left', inplace=True)

        # Modify information about video sources
        video = data['Filename'].str.split('_',  expand=True)[0]
        video = video.astype('category').cat.codes

        # Finalize the dataframe
        df = pd.DataFrame({
            'id': utils.create_id(data['Object ID'].astype(str) + data['Filename']),
            'path': 'data' + os.path.sep + data['Filename'],
            'identity': data['Object ID'],
            'video': video,
            'bbox': bbox,
            'orientation': orientation,
        })
        df = df.join(attributes)
        df.rename({'id': 'image_id'}, axis=1, inplace=True)
        return self.finalize_catalogue(df)



import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN
from tqdm import tqdm
from typing import List, Tuple, Optional
from .lcg import Lcg


class BalancedSplit():
    """Base class for splitting datasets into training and testing sets.

    Implements methods from [this paper](https://arxiv.org/abs/2211.10307).
    Its subclasses need to implement the `split` method.
    It should perform balanced splits separately for all classes.
    Its children are `IdentitySplit` and `TimeAwareSplit`.
    `IdentitySplit` has children `ClosedSetSplit`, `OpenSetSplit` and `DisjointSetSplit`.
    `TimeAwareSplit` has children `TimeProportionSplit` and `TimeCutoffSplit`.
    """

    def __init__(
            self,
            seed: int = 666,
            identity_skip: str = 'unknown',
            col_label: str = 'identity',
            disable_tqdm: bool = True,            
            ) -> None:
        
        self.seed = seed
        self.identity_skip = identity_skip
        self.col_label = col_label
        self.disable_tqdm = disable_tqdm

    def modify_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """Prepares dataframe for splits.

        Removes identities specified in `self.identity_skip` (usually unknown identities).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain columns `identity` and `date`.

        Returns:
            Modified dataframe of the data.
        """
        
        df = df.copy()
        df = df[df[self.col_label] != self.identity_skip]
        return df
    
    def initialize_lcg(self) -> Lcg:
        """Returns the random number generator.

        Returns:
            The random number generator.
        """
        
        return Lcg(self.seed)
    
    def split(self, *args, **kwargs) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Splitting method which needs to be implemented by subclasses.

        It splits the dataframe `df` into labels `idx_train` and `idx_test`.
        The subdataset is obtained by `df.loc[idx_train]` (not `iloc`).

        Returns:
            List of splits. Each split is list of labels of the training and testing sets.
        """

        raise(NotImplementedError('Subclasses should implement this. \n You may want to use ClosedSetSplit instead of BalancedSplit.'))

    def resplit_random(
            self,
            df: pd.DataFrame,
            idx_train: np.ndarray,
            idx_test: np.ndarray
            ) -> Tuple[np.ndarray, np.ndarray]:
        """Creates a random re-split of an already existing split.

        The re-split mimics the split as the training set contains
        the same number of samples for EACH individual.
        The same goes for the testing set.

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain columns `identity` and `date`.
            idx_train (np.ndarray): Labels of the training set.
            idx_test (np.ndarray): Labels of the testing set.

        Returns:
            List of labels of the training and testing sets.
        """

        df = self.modify_df(df)

        # Initialize the random number generator
        lcg = self.initialize_lcg()
        
        # Compute the number of samples for each individual in the training set
        counts_train = {}
        for name, df_name in df.loc[idx_train].groupby(self.col_label):
            counts_train[name] = len(df_name)
        # Compute the number of samples for each individual in the testing set
        counts_test = {}
        for name, df_name in df.loc[idx_test].groupby(self.col_label):
            counts_test[name] = len(df_name)

        idx_train_new = []
        idx_test_new = []
        # Loop over all individuals
        for name, df_name in df.groupby(self.col_label):
            # Extract the number of individuals in the training and testing sets
            n_train = counts_train.get(name, 0)
            n_test = counts_test.get(name, 0)
            if n_train+n_test > 0:
                if len(df_name) < n_train+n_test:
                    raise(Exception('The set is too small.'))
                # Get the correct number of indices in both sets
                idx_permutation = lcg.random_permutation(n_train+n_test)
                idx_permutation = np.array(idx_permutation)
                idx_train_new += list(df_name.index[idx_permutation[:n_train]])
                idx_test_new += list(df_name.index[idx_permutation[n_train:n_train+n_test]])
        return np.array(idx_train_new), np.array(idx_test_new)

    def resplit_by_features(
            self,
            df: pd.DataFrame,
            features: np.ndarray,
            idx_train: np.ndarray,
            save_clusters_prefix: Optional[str] = None,
            **kwargs,
            ) -> Tuple[np.ndarray, np.ndarray]:
        """Creates a random re-split of an already existing split.

        The re-split is based on similarity of features.
        It runs DBSCAN as described in `compute_clusters` and
        performs the clustering as described in `resplit_by_clusters`.
        
        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain column `identity`.
            features (np.ndarray): An array of features with the same length as `df`.
            idx_train (np.ndarray): Labels of the training set.
            save_clusters_prefix (Optional[bool], optional): File name prefix for saving clusters.
            **kwargs (type, optional): See kwargs in `compute_clusters`.

        Returns:
            List of labels of the training and testing sets.
        """        
        
        clusters = self.compute_clusters(df, features, **kwargs)
        if save_clusters_prefix is not None:
            np.save(f'{save_clusters_prefix}.npy', clusters)
        return self.resplit_by_clusters(df, clusters, idx_train)

    def compute_clusters(
            self,
            df: pd.DataFrame,
            features: np.ndarray,
            n_max_cluster: int = 5,
            eps_min: float = 0.01,
            eps_max: float = 0.50,
            eps_step: float = 0.01,
            min_samples: int = 2,
            ) -> np.ndarray:

        """Computes clusters for a random re-split of an already existing split.

        It runs DBSCAN with increasing eps (cluster radius) until
        the clusters are smaller than `n_max_cluster`.
        
        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain column `identity`.
            features (np.ndarray): An array of features with the same length as `df`.
            n_max_cluster (int, optional): Maximal size of cluster before `eps` stops increasing.
            eps_min (float, optional): Lower bound for epsilon.
            eps_max (float, optional): Upper bound for epsilon.
            eps_step (float, optional): Step for epsilon.
            min_samples (int, optional): Minimal cluster size.

        Returns:
            List of clusters.
        """

        df = self.modify_df(df)
        df['cluster'] = np.nan

        for _, df_identity in tqdm(df.groupby(self.col_label), disable=self.disable_tqdm):
            f = features[df.index.get_indexer(df_identity.index)]
            # Run DBScan with increasing eps until there are no clusters bigger than n_max_cluster 
            clusters_saved = None
            for eps in np.arange(eps_min, eps_max+eps_step, eps_step):
                clustering = DBSCAN(eps=eps, min_samples=min_samples)
                clustering.fit(f)
                clusters = pd.Series(clustering.labels_)
                clusters_counts = clusters.value_counts(sort=True)
                # Check if the largest clusters (without outliers) is not too big
                if clusters_counts.index[0] == -1:
                    clustering_failed = len(clusters_counts) > 1 and clusters_counts.iloc[1] > n_max_cluster
                else:
                    clustering_failed = len(clusters_counts) == 1 or clusters_counts.iloc[0] > n_max_cluster
                # If the largest cluster is not too big, save clustering nad continue
                if not clustering_failed:
                    clusters_saved = clusters
                else:
                    break
            
            # Save the clusters
            if clusters_saved is not None:
                clusters_saved[clusters_saved == -1] = np.nan
                df.loc[df_identity.index, 'cluster'] = clusters_saved.to_numpy()

        return df['cluster'].to_numpy()

    def resplit_by_clusters(
            self,
            df: pd.DataFrame,
            clusters: np.ndarray,
            idx_train: np.ndarray,
            ) -> Tuple[np.ndarray, np.ndarray]:
        
        """Creates a random re-split of an already existing split.

        The re-split is based on clusters which collect similar images.
        Then it puts of similar images into the training set.
        The rest is randomly split into training and testing sets.
        The re-split mimics the split as the training set contains
        the same number of samples for EACH individual.
        The same goes for the testing set.
        
        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain column `identity`.
            clusters (np.ndarray): An array of clusters with the same length as `df`.
            idx_train (np.ndarray): Labels of the training set.

        Returns:
            List of labels of the training and testing sets.
        """
        
        df = self.modify_df(df)

        # Replace clusters appearing just ones with np.nan
        clusters_unique, clusters_count = np.unique(clusters, return_counts=True)
        clusters[np.isin(clusters, clusters_unique[clusters_count == 1])] = np.nan
        df['cluster'] = clusters

        # Initialize the random number generator
        lcg = self.initialize_lcg()

        # Determine how many images of each individual should be in the training set
        identity_train_counts = df.loc[idx_train][self.col_label].value_counts()

        # Loop over individuals and create a split for each
        idx_train_new = []
        for identity, df_identity in tqdm(df.groupby(self.col_label), disable=self.disable_tqdm):
            n_train = identity_train_counts.get(identity, 0)
            if len(df_identity) - n_train <= 1:
                # All or all but one samples into the training set
                idx_remaining = np.array(df_identity.index)
                idx_remaining = lcg.random_shuffle(idx_remaining)
                idx_train_identity = idx_remaining[:n_train]
            else:
                # Add all the clusters into the training set
                idx_train_identity = []
                for _, df_cluster in df_identity.groupby('cluster'):
                    # Check if the training set is not too big
                    if len(idx_train_identity) + len(df_cluster) <= n_train:
                        idx_train_identity += list(df_cluster.index)

                # Distribute the remaining indices
                n_train_remaining = n_train - len(idx_train_identity)
                idx_remaining = self.setdiff(df_identity.index, idx_train_identity)
                idx_remaining = lcg.random_shuffle(idx_remaining)
                idx_train_identity += list(idx_remaining[:n_train_remaining])
            idx_train_new += list(idx_train_identity)
        idx_test_new = self.setdiff(df.index, idx_train_new)
        return np.array(idx_train_new), np.array(idx_test_new)

    def set_col_label(self, col_label: str) -> None:
        """Sets col_label to desired value

        Args:
            col_label (str): Desired value for col_label.
        """

        self.col_label = col_label

    def setdiff(self, a, b):
        a = np.array(a)
        b = np.array(b)
        return pd.unique(a[~np.in1d(a,b)])



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .utils import find_images
from .summary import summary

class HolsteinCattleRecognition(DatasetFactory):
    summary = summary['HolsteinCattleRecognition']
    url = 'https://dataverse.nl/api/access/dataset/:persistentId/?persistentId=doi:10.34894/O1ZBSA'
    archive = 'dataset.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
        # Extract all archives in the original archive
        zip_files = find_images('.', img_extensions='zip')
        for _, zip_file in zip_files.iterrows():
            file_name = os.path.join(zip_file['path'], zip_file['file'])
            utils.extract_archive(file_name, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        data = utils.find_images(self.root)

        # Extract the full images only
        idx = data['path'].apply(lambda x: '__MACOSX' not in x)
        data = data[idx]
        idx1 = data['file'].apply(lambda x: os.path.splitext(x)[0].endswith('full photo'))
        idx2 = data['file'].apply(lambda x: os.path.splitext(x)[0].endswith('full photo (2)'))
        data = data[idx1 + idx2]

        folders = data['path'].str.split(os.path.sep, expand=True)
        n_folders = max(folders.columns)

        # Finalize the dataframe
        data['identity'] = folders[n_folders]
        data['path'] = data['path'] + os.path.sep + data['file']
        data['image_id'] = data['file'].str[4:8].astype(str)
        data = data.drop('file', axis=1)

        return self.finalize_catalogue(data)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class MultiCamCows2024(DatasetFactory):
    summary = summary['MultiCamCows2024']
    url = 'https://data.bris.ac.uk/datasets/tar/2inu67jru7a6821kkgehxg3cv2.zip'
    archive = '2inu67jru7a6821kkgehxg3cv2.zip'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
    
    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Finalize the dataframe
        n_cols = len(folders.columns)
        df = pd.DataFrame({
            'image_id': folders[n_cols-1] + '_' + folders[n_cols-2] + '_' + data['file'],
            'path': data['path'] + os.path.sep + data['file'],
            'identity': folders[n_cols-1].astype(int),
            'date': folders[n_cols-2].apply(self.extract_date),
        })
        return self.finalize_catalogue(df)
    
    def extract_date(self, x):
        conversion = {
            'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4,
            'may': 5, 'jun': 6, 'jul': 7, 'aug': 8,
            'sep': 9, 'oct': 10, 'nov': 11, 'dec': 12
            }        
        year = int(x[:4])
        month = conversion[x[4:7].lower()]
        day = int(x[7:])
        return f'{year}-{month}-{day}'



import os
import shutil
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class BirdIndividualID(DatasetFactory):
    summary = summary['BirdIndividualID']
    prefix1 = 'Original_pictures'
    prefix2 = 'IndividualID'
    url = 'https://drive.google.com/uc?id=1YT4w8yF44D-y9kdzgF38z2uYbHfpiDOA'
    archive = 'ferreira_et_al_2020.zip'

    @classmethod
    def _download(cls):
        exception_text = '''Dataset must be downloaded manually.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#birdindividualid'''
        raise Exception(exception_text)
        # utils.gdown_download(cls.url, cls.archive, exception_text=exception_text)
    
    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

        # Create new folder for segmented images
        folder_new = os.getcwd() + 'Segmented'
        if not os.path.exists(folder_new):
            os.makedirs(folder_new)

        # Move segmented images to new folder
        folder_move = 'Cropped_pictures'
        shutil.move(folder_move, os.path.join(folder_new, folder_move))

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        path = os.path.join(self.root, self.prefix1, self.prefix2)
        data = utils.find_images(path)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Remove images with multiple labels
        idx = folders[2].str.contains('_')
        data = data.loc[~idx]
        folders = folders.loc[~idx]

        # Remove some problems with the sociable_weavers/Test_dataset
        if folders.shape[1] == 4:
            idx = folders[3].isnull()
            folders.loc[~idx, 2] = folders.loc[~idx, 3]

        # Extract information from the folder structure
        split = folders[1].replace({'Test_datasets': 'test', 'Test': 'test', 'Train': 'train', 'Val': 'val'})
        identity = folders[2]
        species = folders[0]
        date = data['file'].apply(lambda x: self.extract_date(x))

        # Finalize the dataframe
        df1 = pd.DataFrame({    
            'image_id': utils.create_id(split + data['file']),
            'path': self.prefix1 + os.path.sep + self.prefix2 + os.path.sep + data['path'] + os.path.sep + data['file'],
            'identity': identity,
            'date': date,
            'species': species,
            'original_split': split,
        })

        # Add images without labels
        path = os.path.join(self.root, self.prefix1, 'New_birds')
        data = utils.find_images(path)
        species = data['path']
        date = data['file'].apply(lambda x: self.extract_date(x))

        # Finalize the dataframe
        df2 = pd.DataFrame({    
            'image_id': utils.create_id(data['file']),
            'path': self.prefix1 + os.path.sep + 'New_birds' + os.path.sep + data['path'] + os.path.sep + data['file'],
            'identity': self.unknown_name,
            'date': date,
            'species': species,
            'original_split': np.nan,
        })
        df = pd.concat([df1, df2])
        return self.finalize_catalogue(df)

    def extract_date(self, x):
        x = x.replace(' ', '_')
        date = x.split('_')[-2]
        if date == '2018':
            date = f'2018-{x.split("_")[-3]}-{x.split("_")[-4]}'
        elif date == '1':
            date = x.split('_')[-3]
            date = f'20{date[4:]}-{date[2:4]}-{date[:2]}'
        elif date == 'very':
            date = x.split('_')[-4]
            date = f'20{date[4:]}-{date[2:4]}-{date[:2]}'
        elif date == 'feeder' or date == 'feederB':
            date = f'2018-{x.split("_")[-5]}-{x.split("_")[-4]}'
        elif len(date) == 6 and date != 'feeder':
            date = f'20{date[4:]}-{date[2:4]}-{date[:2]}'
        elif len(date) == 8:
            date = x.split('_')[-3]            
        elif len(date) != 10:
            date = np.nan
        return date

class BirdIndividualIDSegmented(BirdIndividualID):
    prefix1 = 'Cropped_pictures'
    prefix2 = 'IndividuaID'
    warning = '''You are trying to download or extract a segmented dataset.
        It is already included in its non-segmented version.
        Skipping.'''
    
    @classmethod
    def get_data(cls, root, name=None):
        print(cls.warning)

    @classmethod
    def _download(cls):
        print(cls.warning)

    @classmethod
    def _extract(cls):
        print(cls.warning)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class AmvrakikosTurtles(DatasetFactory):    
    summary = summary['AmvrakikosTurtles']
    archive = 'amvrakikosturtles.zip'

    @classmethod
    def _download(cls):
        command = f"datasets download -d wildlifedatasets/amvrakikosturtles --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/preprocessing#amvrakikosturtles'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        data = pd.read_csv(os.path.join(self.root, 'annotations.csv'))

        # Get the bounding box
        columns_bbox = ['bbox_x', 'bbox_y', 'bbox_width', 'bbox_height']
        bbox = data[columns_bbox].to_numpy()
        bbox = pd.Series(list(bbox))
        path = 'images' + os.path.sep + data['image_name']
        date = [utils.get_image_date(os.path.join(self.root, p)) for p in path]

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': range(len(data)),
            'path': path,
            'identity': data['image_name'].apply(lambda x: x.split('_')[0]).astype(int),
            'date': date,
            'orientation': data['image_name'].apply(lambda x: x.split('_')[2]),
            'bbox': bbox,
        })
        df = df[df['orientation'] != 'top']
        df['image_id'] = range(len(df))
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class CatIndividualImages(DatasetFactory):
    summary = summary['CatIndividualImages']
    archive = 'cat-individuals.zip'
    
    @classmethod
    def _download(cls):
        command = f"datasets download -d timost1234/cat-individuals --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#catindividualimages'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)
        
        # Remove 85 duplicate images
        idx = folders[2].isnull()
        data = data[idx]
        folders = folders[idx]

        # Finalize the dataframe
        df = pd.DataFrame({
            'image_id': data['file'].apply(lambda x: os.path.splitext(x)[0]),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': folders[1].astype(int),
        })
        return self.finalize_catalogue(df)



import numpy as np
import pandas as pd

def display_statistics(
        df: pd.DataFrame,
        unknown_name: str = '',
        col_label: str = 'identity',
        ) -> None:
    """Prints statistics about the dataframe.

    Args:
        df (pd.DataFrame): A full dataframe of the data.
        unknown_name (str, optional): Name of the unknown class.
        col_label (str, optional): Column name containing individual animal names (labels).
    """

    # Remove the unknown identities
    df_red = df.loc[df[col_label] != unknown_name, col_label]
    df_red.value_counts().reset_index(drop=True).plot(xlabel='identities', ylabel='counts')
    
    # Compute the total number of identities
    if unknown_name in list(df[col_label].unique()):
        n_identity = len(df.identity.unique()) - 1
    else:
        n_identity = len(df.identity.unique())    
    n_one = len(df.groupby(col_label).filter(lambda x : len(x) == 1))
    n_unidentified = sum(df[col_label] == unknown_name)

    # Print general statistics
    print(f"Number of identitites            {n_identity}")
    print(f"Number of all animals            {len(df)}")
    print(f"Number of animals with one image {n_one}")
    print(f"Number of unidentified animals   {n_unidentified}")

    # Print statistics about video if present
    if 'video' in df.columns:
        print(f"Number of videos                 {len(df[[col_label, 'video']].drop_duplicates())}")
    
    # Print statistics about time span if present
    if 'date' in df.columns:
        span_years = compute_span(df, col_label=col_label) / (60*60*24*365.25)
        if span_years > 1:
            print(f"Images span                      %1.1f years" % (span_years))
        elif span_years / 12 > 1:
            print(f"Images span                      %1.1f months" % (span_years * 12))
        else:
            print(f"Images span                      %1.0f days" % (span_years * 365.25))

def compute_span(df: pd.DataFrame, col_label: str = 'identity') -> float:
    """Compute the time span of the dataset.

    The span is defined as the latest time minus the earliest time of image taken.
    The times are computed separately for each individual.

    Args:
        df (pd.DataFrame): A full dataframe of the data.
        col_label (str, optional): Column name containing individual animal names (labels).

    Returns:
        The span of the dataset in seconds.
    """

    # Convert the dates into timedelta
    df = df.loc[~df['date'].isnull()]
    dates = pd.to_datetime(df['date']).to_numpy()

    # Find the maximal span across individuals
    identities = df[col_label].unique()
    span_seconds = -np.inf
    for identity in identities:
        idx = df[col_label] == identity
        span_seconds = np.maximum(span_seconds, (max(dates[idx]) - min(dates[idx])) / np.timedelta64(1, 's'))
    return span_seconds


import os
import pandas as pd
from . import utils
from .datasets_wildme import DatasetFactoryWildMe
from .summary import summary

class GiraffeZebraID(DatasetFactoryWildMe):
    summary = summary['GiraffeZebraID']
    url = 'https://lilawildlife.blob.core.windows.net/lila-wildlife/wild-me/gzgc.coco.tar.gz'
    archive = 'gzgc.coco.tar.gz'

    @classmethod
    def _download(cls):
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)
    
    def create_catalogue(self) -> pd.DataFrame:
        return self.create_catalogue_wildme('gzgc', 2020)



import os
import numpy as np
import pandas as pd
from PIL import Image
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class PrimFace(DatasetFactory):
    summary = summary['PrimFace']
    downloads = [
        ('https://visiome.neuroinf.jp/database/file/1508/j01.zip', 'j01.zip'),
        ('https://visiome.neuroinf.jp/database/file/1518/j02.zip', 'j02.zip'),
        ('https://visiome.neuroinf.jp/database/file/1525/j03.zip', 'j03.zip'),
        ('https://visiome.neuroinf.jp/database/file/1530/j04.zip', 'j04.zip'),
        ('https://visiome.neuroinf.jp/database/file/1539/j05.zip', 'j05.zip'),
        ('https://visiome.neuroinf.jp/database/file/1504/j06.zip', 'j06.zip'),
        ('https://visiome.neuroinf.jp/database/file/1513/j07.zip', 'j07.zip'),
        ('https://visiome.neuroinf.jp/database/file/1553/j08.zip', 'j08.zip'),
        ('https://visiome.neuroinf.jp/database/file/1612/j09.zip', 'j09.zip'),
        ('https://visiome.neuroinf.jp/database/file/1555/j10.zip', 'j10.zip'),
        ('https://visiome.neuroinf.jp/database/file/1557/j11.zip', 'j11.zip'),
        ('https://visiome.neuroinf.jp/database/file/1559/j12.zip', 'j12.zip'),
        ('https://visiome.neuroinf.jp/database/file/1561/j13.zip', 'j13.zip'),
        ('https://visiome.neuroinf.jp/database/file/1516/j14.zip', 'j14.zip'),
        ('https://visiome.neuroinf.jp/database/file/1522/j15.zip', 'j15.zip'),
        ('https://visiome.neuroinf.jp/database/file/1526/j16.zip', 'j16.zip'),
        ('https://visiome.neuroinf.jp/database/file/1531/j17.zip', 'j17.zip'),
        ('https://visiome.neuroinf.jp/database/file/1535/j18.zip', 'j18.zip'),
        ('https://visiome.neuroinf.jp/database/file/1573/j19.2.zip', 'j19.2.zip'),
        ('https://visiome.neuroinf.jp/database/file/1503/r01.zip', 'r01.zip'),
        ('https://visiome.neuroinf.jp/database/file/1509/r02.zip', 'r02.zip'),
        ('https://visiome.neuroinf.jp/database/file/1515/r03.zip', 'r03.zip'),
        ('https://visiome.neuroinf.jp/database/file/1521/r04.zip', 'r04.zip'),
        ('https://visiome.neuroinf.jp/database/file/1615/r05.zip', 'r05.zip'),
        ('https://visiome.neuroinf.jp/database/file/1617/r06.zip', 'r06.zip'),
        ('https://visiome.neuroinf.jp/database/file/1619/r07.zip', 'r07.zip'),
        ('https://visiome.neuroinf.jp/database/file/1621/r08.zip', 'r08.zip'),
        ('https://visiome.neuroinf.jp/database/file/1623/r09.zip', 'r09.zip'),
        ('https://visiome.neuroinf.jp/database/file/1627/r10.zip', 'r10.zip'),
        ('https://visiome.neuroinf.jp/database/file/1629/r11.zip', 'r11.zip'),
        ('https://visiome.neuroinf.jp/database/file/1633/r12.zip', 'r12.zip'),
        ('https://visiome.neuroinf.jp/database/file/1635/r13.zip', 'r13.zip'),
        ('https://visiome.neuroinf.jp/database/file/1637/r14.zip', 'r14.zip'),
        ('https://visiome.neuroinf.jp/database/file/1639/r15.zip', 'r15.zip'),
        ('https://visiome.neuroinf.jp/database/file/1641/r16.zip', 'r16.zip'),
        ('https://visiome.neuroinf.jp/database/file/1643/r17.zip', 'r17.zip'),
        ('https://visiome.neuroinf.jp/database/file/1645/r18.zip', 'r18.zip'),
        ('https://visiome.neuroinf.jp/database/file/1647/r19.zip', 'r19.zip'),
        ('https://visiome.neuroinf.jp/database/file/1649/r20.zip', 'r20.zip'),
        ('https://visiome.neuroinf.jp/database/file/1651/r21.zip', 'r21.zip'),
        ('https://visiome.neuroinf.jp/database/file/1653/r22.zip', 'r22.zip'),
        ('https://visiome.neuroinf.jp/database/file/1655/r23.zip', 'r23.zip'),
        ('https://visiome.neuroinf.jp/database/file/1657/r24.zip', 'r24.zip'),
        ('https://visiome.neuroinf.jp/database/file/1659/r25.zip', 'r25.zip'),
        ('https://visiome.neuroinf.jp/database/file/1661/r26.zip', 'r26.zip'),
        ('https://visiome.neuroinf.jp/database/file/1663/r27.zip', 'r27.zip'),
        ('https://visiome.neuroinf.jp/database/file/1665/r28.zip', 'r28.zip'),
        ('https://visiome.neuroinf.jp/database/file/1667/r29.zip', 'r29.zip'),
        ('https://visiome.neuroinf.jp/database/file/1669/r30.zip', 'r30.zip'),
        ('https://visiome.neuroinf.jp/database/file/1671/r31.zip', 'r31.zip'),
        ('https://visiome.neuroinf.jp/database/file/1673/r32.zip', 'r32.zip'),
        ('https://visiome.neuroinf.jp/database/file/1675/r33.zip', 'r33.zip'),
        ('https://visiome.neuroinf.jp/database/file/1677/r34.zip', 'r34.zip'),
        ('https://visiome.neuroinf.jp/database/file/1679/r35.zip', 'r35.zip'),
        ('https://visiome.neuroinf.jp/database/file/1681/r36.zip', 'r36.zip'),
        ('https://visiome.neuroinf.jp/database/file/1683/r37.zip', 'r37.zip'),
        ('https://visiome.neuroinf.jp/database/file/1685/r38.zip', 'r38.zip'),
        ('https://visiome.neuroinf.jp/database/file/1563/c01.zip', 'c01.zip'),
        ('https://visiome.neuroinf.jp/database/file/1565/c02.zip', 'c02.zip'),
        ('https://visiome.neuroinf.jp/database/file/1567/c03.zip', 'c03.zip'),
        ('https://visiome.neuroinf.jp/database/file/1569/c04.zip', 'c04.zip'),
        ('https://visiome.neuroinf.jp/database/file/1571/c05.zip', 'c05.zip'),
        ('https://visiome.neuroinf.jp/database/file/1691/c06.zip', 'c06.zip'),
        ('https://visiome.neuroinf.jp/database/file/1693/c07.zip', 'c07.zip'),
        ('https://visiome.neuroinf.jp/database/file/1695/c08.zip', 'c08.zip'),
        ('https://visiome.neuroinf.jp/database/file/1697/c09.zip', 'c09.zip'),
        ('https://visiome.neuroinf.jp/database/file/1699/c10.zip', 'c10.zip'),
        ('https://visiome.neuroinf.jp/database/file/1701/c11.zip', 'c11.zip'),
    ]

    @classmethod
    def _download(cls):
        for url, archive in cls.downloads:
            utils.download_url(url, archive)

    @classmethod
    def _extract(cls):
        for url, archive in cls.downloads:
            utils.extract_archive(archive, extract_path=os.path.splitext(archive)[0], delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        
        # Finalize the dataframe
        species = {'c': 'chimpanzee', 'j': 'japanese monkey', 'r': 'rhesus monkey'}
        df = pd.DataFrame({
            'image_id': data['file'].apply(lambda x: os.path.splitext(x)[0]),
            'path': data['path'] + os.path.sep + data['file'],
            'identity': data['file'].apply(lambda x: x.split('_')[0]),
            'species': data['file'].apply(lambda x: species[x[0]]),
        })
        return self.finalize_catalogue(df)

    def load_image(self, path: str) -> Image:
        """Load an image with `path`. Need to write this because PIL loads (w,h,4), where the last channel is transparency.

        Args:
            path (str): Path to the image.

        Returns:
            Loaded image.
        """

        img = Image.open(path)
        return Image.fromarray(np.array(img)[:,:,:3])



from .datasets import DatasetFactory, WildlifeDataset
from .aau_zebrafish import AAUZebraFish
from .aerial_cattle import AerialCattle2017
from .amvrakikos_turtles import AmvrakikosTurtles
from .animal_clef import AnimalCLEF2025, AnimalCLEF2025_LynxID2025, AnimalCLEF2025_SalamanderID2025, AnimalCLEF2025_SeaTurtleID2022
from .atrw import ATRW
from .beluga_id import BelugaID, BelugaIDv2
from .bird_individual_id import BirdIndividualID, BirdIndividualIDSegmented
from .bristol_gorillas_2020 import BristolGorillas2020
from .cattle_muzzle import CattleMuzzle
from .cat_individual_images import CatIndividualImages
from .cobra_re_identification_youngstock import CoBRAReIdentificationYoungstock
from .ctai import CTai
from .czoo import CZoo
from .chicks4free_id import Chicks4FreeID
from .cow_dataset import CowDataset
from .cows import Cows2021, Cows2021v2
from .czechlynx import CzechLynx
from .dog_face_net import DogFaceNet
from .drosophila import Drosophila
from .elpephants import ELPephants
from .friesian_cattle import FriesianCattle2015, FriesianCattle2015v2, FriesianCattle2017
from .giraffe_zebra_id import GiraffeZebraID
from .giraffes import Giraffes
from .happy_whale import HappyWhale
from .holstein_cattle_recognition import HolsteinCattleRecognition
from .humpback_whale_id import HumpbackWhaleID
from .hyena_id import HyenaID2022
from .ipanda import IPanda50
from .leopard_id import LeopardID2022
from .lion_data import LionData
from .macaque_faces import MacaqueFaces
from .mpdd import MPDD
from .multi_cam_cows import MultiCamCows2024
from .ndd import NDD20, NDD20v2
from .noaa_right_whale import NOAARightWhale
from .nyala_data import NyalaData
from .open_cows import OpenCows2020
from .polar_bear_vid_id import PolarBearVidID
from .prim_face import PrimFace
from .reunion_turtles import ReunionTurtles
from .seal_id import SealID, SealIDSegmented
from .sea_star_reid import SeaStarReID2023
from .sea_turtle_id import SeaTurtleID2022, SeaTurtleIDHeads
from .smalst import SMALST
from .southern_province_turtles import SouthernProvinceTurtles
from .stripe_spotter import StripeSpotter
from .wildlife_reid_10k import WildlifeReID10k
from .wild_raptor_id import WildRaptorID
from .whaleshark_id import WhaleSharkID
from .zakynthos_turtles import ZakynthosTurtles
from .zindi_turtle_recall import ZindiTurtleRecall
from .summary import Summary
from .utils import get_image, load_image

names_all = [
    AAUZebraFish,
    AerialCattle2017,
    AmvrakikosTurtles,
    ATRW,
    BelugaIDv2,
    BirdIndividualID, BirdIndividualIDSegmented,
    BristolGorillas2020,
    CattleMuzzle,
    CatIndividualImages,
    Chicks4FreeID,
    CoBRAReIdentificationYoungstock,
    CowDataset,
    Cows2021v2,
    CzechLynx,
    CTai,
    CZoo,
    DogFaceNet,
    Drosophila,
    ELPephants,
    FriesianCattle2015v2,
    FriesianCattle2017,
    GiraffeZebraID,
    Giraffes,
    HappyWhale,
    HolsteinCattleRecognition,
    HumpbackWhaleID,
    HyenaID2022,
    IPanda50,
    LeopardID2022,
    LionData,
    MacaqueFaces,
    MPDD,
    MultiCamCows2024,
    NDD20v2,
    NOAARightWhale,
    NyalaData,
    OpenCows2020,
    PolarBearVidID,
    PrimFace,
    ReunionTurtles,
    SealID, SealIDSegmented,
    SeaStarReID2023,
    SeaTurtleID2022, SeaTurtleIDHeads,
    SMALST,
    SouthernProvinceTurtles,
    StripeSpotter,
    WildRaptorID,
    WhaleSharkID,
    ZakynthosTurtles,
    ZindiTurtleRecall,
]

names_wild = [
    AmvrakikosTurtles,
    BelugaIDv2,
    CzechLynx,
    ELPephants,
    GiraffeZebraID,
    HappyWhale,
    HumpbackWhaleID,
    HyenaID2022,
    LeopardID2022,
    NDD20v2,
    NOAARightWhale,
    NyalaData,
    ReunionTurtles,
    SealID,
    SeaTurtleID2022,
    SouthernProvinceTurtles,
    StripeSpotter,
    WildRaptorID,
    WhaleSharkID,
    ZakynthosTurtles,
]

names_small = [
    AerialCattle2017,
    BelugaIDv2,
    CoBRAReIdentificationYoungstock,
    CTai,
    CZoo,
    DogFaceNet,
    ELPephants,
    FriesianCattle2015v2,
    FriesianCattle2017,
    HolsteinCattleRecognition,
    IPanda50,
    MacaqueFaces,
    MPDD,
    NyalaData,
    PolarBearVidID,
    ReunionTurtles,
    SeaTurtleIDHeads,
    SouthernProvinceTurtles,
    StripeSpotter,
    ZakynthosTurtles,
]

names_cows = [
    AerialCattle2017,
    CattleMuzzle,
    CoBRAReIdentificationYoungstock,
    CowDataset,
    Cows2021v2,
    FriesianCattle2015v2,
    FriesianCattle2017,
    HolsteinCattleRecognition,
    MultiCamCows2024,
    OpenCows2020,
]

names_dogs = [
    DogFaceNet,
    MPDD,
]

names_giraffes = [
    GiraffeZebraID,
    Giraffes,
    SMALST,
    StripeSpotter,
]

names_primates = [
    BristolGorillas2020,
    CTai,
    CZoo,
    MacaqueFaces,
    PrimFace,
]

names_turtles = [
    AmvrakikosTurtles,
    ReunionTurtles,
    SeaTurtleIDHeads,
    SouthernProvinceTurtles,
    ZakynthosTurtles,
    ZindiTurtleRecall,
]

names_whales = [
    BelugaIDv2,
    HappyWhale,
    HumpbackWhaleID,
    NDD20v2,
    NOAARightWhale,
    WhaleSharkID,
]



import os
import json
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class SeaTurtleID2022(DatasetFactory):
    summary = summary['SeaTurtleID2022']
    archive = 'seaturtleid2022.zip'

    @classmethod
    def _download(cls):
        command = f"datasets download -d wildlifedatasets/seaturtleid2022 --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#seaturtleid2022'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self, category_name='head') -> pd.DataFrame:
        """Creates dataframe for SeaTurtleID2022.

        Args:
            category_name (str, optional): Choose one from ['turtle', 'flipper', 'head'].

        Returns:
            Created dataframe.
        """

        # Load annotations JSON file
        path_json = os.path.join('turtles-data', 'data', 'annotations.json')
        with open(os.path.join(self.root, path_json)) as file:
            data = json.load(file)
        path_csv = os.path.join('turtles-data', 'data', 'metadata_splits.csv')
        with open(os.path.join(self.root, path_csv)) as file:
            df_images = pd.read_csv(file)
        # Extract categories
        categories = {}
        for category in data['categories']:
            categories[category['name']] = category['id']
        if category_name not in categories:
            #printfor category in categories.keys()
            raise Exception(f'Category {category_name} not allowed. Choose one from {categories.keys()}.')
        category_id = categories[category_name]

        # Extract data from the JSON file
        create_dict = lambda i: {
            'id': i['id'],
            'bbox': i['bbox'],
            'image_id': i['image_id'],
            'segmentation': i['segmentation'],
            'orientation': i['attributes']['orientation'] if 'orientation' in i['attributes'] else np.nan
        }
        df_annotation = pd.DataFrame([create_dict(i) for i in data['annotations'] if i['category_id'] == category_id])
        idx_bbox = ~df_annotation['bbox'].isnull()
        df_annotation.loc[idx_bbox,'bbox'] = df_annotation.loc[idx_bbox,'bbox'].apply(lambda x: eval(x) if isinstance(x, str) else x)
        df_images.rename({'id': 'image_id'}, axis=1, inplace=True)

        # Merge the information from the JSON file
        df = pd.merge(df_images, df_annotation, on='image_id', how='outer')
        df['path'] = 'turtles-data' + os.path.sep + 'data' + os.path.sep + df['file_name'].str.replace('/', os.path.sep)
        df = df.drop(['id', 'file_name', 'timestamp', 'width', 'height', 'year', 'split_closed_random', 'split_open'], axis=1)
        df.rename({'split_closed': 'original_split'}, axis=1, inplace=True)
        df['date'] = df['date'].apply(lambda x: x[:4] + '-' + x[5:7] + '-' + x[8:10])

        df['image_id'] = range(1, len(df)+1)
        return self.finalize_catalogue(df)

class SeaTurtleIDHeads(DatasetFactory):
    summary = summary['SeaTurtleIDHeads']
    archive = 'seaturtleidheads.zip'

    @classmethod
    def _download(cls):
        command = f"datasets download -d wildlifedatasets/seaturtleidheads --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#seaturtleid'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Load annotations JSON file
        path_json = 'annotations.json'
        with open(os.path.join(self.root, path_json)) as file:
            data = json.load(file)

        # Extract dtaa from the JSON file
        create_dict = lambda i: {'id': i['id'], 'image_id': i['image_id'], 'identity': i['identity'], 'orientation': i['position']}
        df_annotation = pd.DataFrame([create_dict(i) for i in data['annotations']])
        create_dict = lambda i: {'file_name': i['path'].split('/')[-1], 'image_id': i['id'], 'date': i['date']}
        df_images = pd.DataFrame([create_dict(i) for i in data['images']])

        # Merge the information from the JSON file
        df = pd.merge(df_annotation, df_images, on='image_id')
        df['path'] = 'images' + os.path.sep + df['identity'] + os.path.sep + df['file_name']        
        df = df.drop(['image_id', 'file_name'], axis=1)
        df['date'] = df['date'].apply(lambda x: x[:4] + '-' + x[5:7] + '-' + x[8:10])

        # Finalize the dataframe
        df.rename({'id': 'image_id'}, axis=1, inplace=True)
        return self.finalize_catalogue(df)



import os
import json
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class ATRW(DatasetFactory):
    summary = summary['ATRW']
    url = 'https://github.com/cvwc2019/ATRWEvalScript/archive/refs/heads/main.zip'
    archive = 'main.zip'
    downloads = [
        # Wild dataset (Detection)
        ('https://lilawildlife.blob.core.windows.net/lila-wildlife/cvwc2019/test/atrw_detection_test.tar.gz', 'atrw_detection_test.tar.gz'),

        # Re-ID dataset
        ('https://lilawildlife.blob.core.windows.net/lila-wildlife/cvwc2019/train/atrw_reid_train.tar.gz', 'atrw_reid_train.tar.gz'),
        ('https://lilawildlife.blob.core.windows.net/lila-wildlife/cvwc2019/train/atrw_anno_reid_train.tar.gz', 'atrw_anno_reid_train.tar.gz'),
        ('https://lilawildlife.blob.core.windows.net/lila-wildlife/cvwc2019/test/atrw_reid_test.tar.gz', 'atrw_reid_test.tar.gz'),
        ('https://lilawildlife.blob.core.windows.net/lila-wildlife/cvwc2019/test/atrw_anno_reid_test.tar.gz', 'atrw_anno_reid_test.tar.gz'),
    ]

    @classmethod
    def _download(cls):
        for url, archive in cls.downloads:
            utils.download_url(url, archive)
        # Evaluation scripts
        utils.download_url(cls.url, cls.archive)

    @classmethod
    def _extract(cls):
        for url, archive in cls.downloads:
            archive_name = archive.split('.')[0]
            utils.extract_archive(archive, archive_name, delete=True)
        # Evaluation scripts
        utils.extract_archive(cls.archive, 'eval_script', delete=True)
    
    def create_catalogue(self) -> pd.DataFrame:
        # Load information for the reid_train part of the dataset
        ids = pd.read_csv(os.path.join(self.root, 'atrw_anno_reid_train', 'reid_list_train.csv'),
                        names=['identity', 'path'],
                        header=None
                        )
        ids['id'] = ids['path'].str.split('.', expand=True)[0].astype(int)
        ids['original_split'] = 'train'

        # Load keypoints for the reid_train part of the dataset
        with open(os.path.join(self.root, 'atrw_anno_reid_train', 'reid_keypoints_train.json')) as file:
            keypoints = json.load(file)
        df_keypoints = {
            'path': pd.Series(keypoints.keys()),
            'keypoints': pd.Series(list(pd.DataFrame([keypoints[key] for key in keypoints.keys()]).to_numpy())),
        }
        data = pd.DataFrame(df_keypoints)

        # Merge information for the reid_train part of the dataset
        df_train = pd.merge(ids, data, on='path', how='left')
        df_train['path'] = 'atrw_reid_train' + os.path.sep + 'train' + os.path.sep + df_train['path']

        # Load information for the test_plain part of the dataset
        with open(os.path.join(self.root, 'eval_script', 'ATRWEvalScript-main', 'annotations', 'gt_test_plain.json')) as file:
            identity = json.load(file)
        identity = pd.DataFrame(identity)
        ids = pd.read_csv(os.path.join(self.root, 'atrw_anno_reid_test', 'reid_list_test.csv'),
                        names=['path'],
                        header=None
                        )
        ids['id'] = ids['path'].str.split('.', expand=True)[0].astype(int)
        ids['original_split'] = 'test'
        ids = pd.merge(ids, identity, left_on='id', right_on='imgid', how='left')
        ids = ids.drop(['query', 'frame', 'imgid'], axis=1)
        ids.rename(columns = {'entityid': 'identity'}, inplace = True)

        # Load keypoints for the test part of the dataset
        with open(os.path.join(self.root, 'atrw_anno_reid_test', 'reid_keypoints_test.json')) as file:
            keypoints = json.load(file)
        df_keypoints = {
            'path': pd.Series(keypoints.keys()),
            'keypoints': pd.Series(list(pd.DataFrame([keypoints[key] for key in keypoints.keys()]).to_numpy())),
        }
        data = pd.DataFrame(df_keypoints)

        # Merge information for the test_plain part of the dataset
        df_test1 = pd.merge(ids, data, on='path', how='left')
        df_test1['path'] = 'atrw_reid_test' + os.path.sep + 'test' + os.path.sep + df_test1['path']

        # Load information for the test_wild part of the dataset
        with open(os.path.join(self.root, 'eval_script', 'ATRWEvalScript-main', 'annotations', 'gt_test_wild.json')) as file:
            identity = json.load(file)
        ids = utils.find_images(os.path.join(self.root, 'atrw_detection_test', 'test'))
        ids['imgid'] = ids['file'].str.split('.', expand=True)[0].astype('int')
        entries = []
        for key in identity.keys():
            for entry in identity[key]:
                bbox = [entry['bbox'][0], entry['bbox'][1], entry['bbox'][2]-entry['bbox'][0], entry['bbox'][3]-entry['bbox'][1]]
                entries.append({'imgid': int(key), 'bbox': bbox, 'identity': entry['eid']})
        entries = pd.DataFrame(entries)

        # Merge information for the test_wild part of the dataset
        df_test2 = pd.merge(ids, entries, on='imgid', how='left')
        df_test2['path'] = 'atrw_detection_test' + os.path.sep + 'test' + os.path.sep + df_test2['file']
        df_test2['id'] = df_test2['imgid'].astype(str) + '_' + df_test2['identity'].astype(str)
        df_test2['original_split'] = 'test'
        df_test2 = df_test2.drop(['file', 'imgid'], axis=1)

        # Finalize the dataframe
        df = pd.concat([df_train, df_test1, df_test2])
        df['id'] = utils.create_id(df['id'].astype(str))
        df.rename({'id': 'image_id'}, axis=1, inplace=True)
        return self.finalize_catalogue(df)



import numpy as np
import pandas as pd
from typing import List, Tuple
from .balanced_split import BalancedSplit


class FullSplit(BalancedSplit):
    """Simplest split returning the whole dataset.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def split(self, df: pd.DataFrame) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Implementation of the [base splitting method](../reference_splits#splits.balanced_split.BalancedSplit.split).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain column `identity`.

        Returns:
            List of splits. Each split is list of labels of the training and testing sets.
        """

        return [ (df.index.values, np.array([], dtype=int)) ]
    

class IdentitySplit(BalancedSplit):
    """Base class for `ClosedSetSplit`, `OpenSetSplit` and `DisjointSetSplit`.
    """
    
    def general_split(
            self,
            df: pd.DataFrame,
            individual_train: List[str],
            individual_test: List[str],
            ) -> Tuple[np.ndarray, np.ndarray]:
        """General-purpose split into the training and testing sets.

        It puts all samples of `individual_train` into the training set
        and all samples of `individual_test` into the testing set.
        The splitting is performed for each individual separately.
        The split will result in at least one sample in both the training and testing sets.
        If only one sample is available for an individual, it will be in the training set.
                
        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain column `identity`.
            individual_train (List[str]): Individuals to be only in the training test.
            individual_test (List[str]): Individuals to be only in the testing test.

        Returns:
            List of labels of the training and testing sets.
        """
        
        # Initialize the random number generator
        lcg = self.initialize_lcg()

        # Compute how many samples go automatically to the training and testing sets
        y_counts = df[self.col_label].value_counts()
        n_train = sum([y_counts.loc[y] for y in individual_train])
        n_test = sum([y_counts.loc[y] for y in individual_test])
        
        # Recompute ratio_train and adjust it to proper bounds
        n = len(df)
        ratio_train = self.ratio_train
        if n_train + n_test > 0 and n_train + n_test < n:
            ratio_train = (n*ratio_train - n_train) / (n - n_test - n_train)
        ratio_train = np.clip(ratio_train, 0, 1)

        idx_train = []
        idx_test = []        
        # Make a loop over all individuals
        for individual, df_individual in df.groupby(self.col_label):
            if individual in individual_train and individual in individual_test:
                # Check if the class does not belong to both sets
                raise(Exception('Individual cannot be both in individual_train and individual_test.'))
            elif individual in individual_train:
                # Check if the class does not belong to the training set
                idx_train += list(df_individual.index)
            elif individual in individual_test:
                # Check if the class does not belong to the testing set
                idx_test += list(df_individual.index)
            else:
                # Otherwise compute the number of samples in the training set
                n_individual = len(df_individual)
                n_train = np.round(ratio_train * n_individual).astype(int)
                if n_train == n_individual and n_train > 1:
                    n_train -= 1
                if n_train == 0:
                    n_train = 1
                # Create indices to the training set and randomly permute them                
                idx_permutation = lcg.random_permutation(n_individual)
                idx_permutation = np.array(idx_permutation)
                idx_train += list(df_individual.index[idx_permutation[:n_train]])
                idx_test += list(df_individual.index[idx_permutation[n_train:]])
        return np.array(idx_train), np.array(idx_test)
    

class ClosedSetSplit(IdentitySplit):
    """Closed-set splitting method into training and testing sets.

    All individuals are both in the training and testing set.
    The only exception is that individuals with only one sample are in the training set.
    Implementation of [this paper](https://arxiv.org/abs/2211.10307).
    """

    def __init__(
            self,
            ratio_train: float,
            **kwargs
            ) -> None:
        """Initializes the class.

        Args:
            ratio_train (float): *Approximate* size of the training set.
            **kwargs (type, optional): See kwargs `seed`, `identity_skip` and `col_label` of the parent class.
        """

        self.ratio_train = ratio_train
        super().__init__(**kwargs)
    
    def split(self, df: pd.DataFrame) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Implementation of the [base splitting method](../reference_splits#splits.balanced_split.BalancedSplit.split).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain column `identity`.

        Returns:
            List of splits. Each split is list of labels of the training and testing sets.
        """

        df = self.modify_df(df)
        individual_train = np.array([], dtype=object)
        individual_test = np.array([], dtype=object)
        return [self.general_split(df, individual_train, individual_test)]


class OpenSetSplit(IdentitySplit):
    """Open-set splitting method into training and testing sets.

    Some individuals are in the testing but not in the training set.
    Implementation of [this paper](https://arxiv.org/abs/2211.10307).
    """

    def __init__(
            self,
            ratio_train: float,
            ratio_class_test: float = None,
            n_class_test: int = None,
            open_in_test: bool = True,
            **kwargs
            ) -> None:
        """Initializes the class.

        The user must provide exactly one from `ratio_class_test` and `n_class_test`.
        The latter specifies the number of individuals to be only in the testing set.
        The former specified the ratio of samples of individuals (not individuals themselves)
        to be only in the testing set.

        Args:
            ratio_train (float): *Approximate* size of the training set.
            ratio_class_test (float, optional): *Approximate* ratio of samples of individuals only in the testing set.
            n_class_test (int, optional): Number of individuals only in the testing set.
            open_in_test (str, optional): Whether the unique identifies will be in test (default) or train set.
            **kwargs (type, optional): See kwargs `seed`, `identity_skip` and `col_label` of the parent class.
        """

        if ratio_class_test is None and n_class_test is None:
            raise(Exception('Either ratio_class_test or n_class_test must be provided.'))
        elif ratio_class_test is not None and n_class_test is not None:
            raise(Exception('Only ratio_class_test or n_class_test can be provided.'))
        
        self.ratio_train = ratio_train
        self.ratio_class_test = ratio_class_test
        self.n_class_test = n_class_test
        self.open_in_test = open_in_test
        super().__init__(**kwargs)

    def split(self, df: pd.DataFrame) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Implementation of the [base splitting method](../reference_splits#splits.balanced_split.BalancedSplit.split).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain column `identity`.

        Returns:
            List of splits. Each split is list of labels of the training and testing sets.
        """
        
        df = self.modify_df(df)

        # Initialize the random number generator
        lcg = self.initialize_lcg()

        # Compute the counts and randomly permute them
        y_counts = df[self.col_label].value_counts()
        n_class = len(y_counts)
        idx = lcg.random_permutation(n_class)
        y_counts = y_counts.iloc[idx]

        # Compute number of identities in the testing set
        n = len(df)
        if self.n_class_test is None:
            n_test = np.round(n*self.ratio_class_test).astype(int)
            n_class_test = np.where(np.cumsum(y_counts) >= n_test)[0][0] + 1
        else:
            n_class_test = self.n_class_test

        # Specify individuals going purely into training and testing sets
        if self.open_in_test:
            individual_train = np.array([], dtype=object)
            individual_test = np.array(y_counts.index[:n_class_test])
        else:
            individual_train = np.array(y_counts.index[:n_class_test])
            individual_test = np.array([], dtype=object)
        return [self.general_split(df, individual_train, individual_test)]


class DisjointSetSplit(IdentitySplit):
    """Disjoint-set splitting method into training and testing sets.

    No individuals are in both the training and testing sets.
    Implementation of [this paper](https://arxiv.org/abs/2211.10307).
    """

    def __init__(
            self,
            ratio_class_test: float = None,
            n_class_test: int = None,
            **kwargs
            ) -> None:
        """Initializes the class.

        The user must provide exactly one from `ratio_class_test` and `n_class_test`.
        The latter specifies the number of individuals to be only in the testing set.
        The former specified the ratio of samples of individuals (not individuals themselves)
        to be only in the testing set.

        Args:
            ratio_class_test (float, optional): *Approximate* ratio of samples of individuals only in the testing set.
            n_class_test (int, optional): Number of individuals only in the testing set.
            **kwargs (type, optional): See kwargs `seed`, `identity_skip` and `col_label` of the parent class.
        """

        if ratio_class_test is None and n_class_test is None:
            raise(Exception('Either ratio_class_test or n_class_test must be provided.'))
        elif ratio_class_test is not None and n_class_test is not None:
            raise(Exception('Only ratio_class_test or n_class_test can be provided.'))
        
        self.ratio_train = 0 # Arbitrary value
        self.ratio_class_test = ratio_class_test
        self.n_class_test = n_class_test
        super().__init__(**kwargs)

    def split(self, df: pd.DataFrame) -> List[Tuple[np.ndarray, np.ndarray]]:
        """Implementation of the [base splitting method](../reference_splits#splits.balanced_split.BalancedSplit.split).

        Args:
            df (pd.DataFrame): A dataframe of the data. It must contain column `identity`.
        
        Returns:
            List of splits. Each split is list of labels of the training and testing sets.
        """

        df = self.modify_df(df)
        
        # Initialize the random number generator
        lcg = self.initialize_lcg()

        # Compute the counts and randomly permute them
        y_counts = df[self.col_label].value_counts()
        n_class = len(y_counts)
        idx = lcg.random_permutation(n_class)
        y_counts = y_counts.iloc[idx]

        # Compute number of identities in the testing set
        n = len(df)
        if self.n_class_test is None:
            n_test = np.round(n*self.ratio_class_test).astype(int)
            n_class_test = np.where(np.cumsum(y_counts) >= n_test)[0][0] + 1
        else:
            n_class_test = self.n_class_test

        # Specify individuals going purely into training and testing sets
        individual_train = np.array(y_counts.index[n_class_test:])
        individual_test = np.array(y_counts.index[:n_class_test])
        return [self.general_split(df, individual_train, individual_test)]




import os
import shutil
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class SMALST(DatasetFactory):
    summary = summary['SMALST']
    url = 'https://drive.google.com/uc?id=1yVy4--M4CNfE5x9wUr1QBmAXEcWb6PWF'
    archive = 'zebra_training_set.zip'

    @classmethod
    def _download(cls):
        exception_text = '''Download failed. GDown quota probably reached. Download dataset manually.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#smalst'''
        utils.gdown_download(cls.url, cls.archive, exception_text)

    @classmethod
    def _extract(cls):
        exception_text = '''Extracting works only on Linux. Please extract it manually.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#smalst'''
        if os.name == 'posix':
            os.system('jar xvf ' + cls.archive)
            os.remove(cls.archive)
            shutil.rmtree(os.path.join('zebra_training_set', 'annotations'))
            shutil.rmtree(os.path.join('zebra_training_set', 'texmap'))
            shutil.rmtree(os.path.join('zebra_training_set', 'uvflow'))

        else:
            raise Exception(exception_text)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(os.path.join(self.root, 'zebra_training_set', 'images'))
        
        # Extract information about the images
        path = data['file'].str.strip('zebra_')
        data['identity'] = path.str[0]
        data['image_id'] = [int(p[1:].strip('_frame_').split('.')[0]) for p in path]
        data['path'] = 'zebra_training_set' + os.path.sep + 'images' + os.path.sep + data['file']
        data = data.drop(['file'], axis=1)

        # Find all masks in root
        masks = utils.find_images(os.path.join(self.root, 'zebra_training_set', 'bgsub'))
        
        # Extract information about the images
        path = masks['file'].str.strip('zebra_')
        masks['image_id'] = [int(p[1:].strip('_frame_').split('.')[0]) for p in path]
        masks['segmentation'] = 'zebra_training_set' + os.path.sep + 'bgsub' + os.path.sep + masks['file']
        masks = masks.drop(['path', 'file'], axis=1)

        # Finalize the dataframe
        df = pd.merge(data, masks, on='image_id', how='left')
        return self.finalize_catalogue(df)



import os
import pandas as pd
import numpy as np
from typing import Tuple, List
import hashlib
import urllib.request
from tqdm import tqdm
import shutil
from contextlib import contextmanager
from PIL import Image, ImageOps
import cv2

def load_image(path: str, max_size: int = None) -> Image:
    """Loads an image.

    Args:
        path (str): Path of the image.
        max_size (int, optional): Maximal size of the image or None (no restriction).

    Returns:
        Loaded image.
    """

    # We load it with OpenCV because PIL does not apply metadata.
    img = cv2.imread(path)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img = Image.fromarray(img)
    if max_size is not None:
        w, h = img.size
        if max(w, h) > max_size:
            c = max_size / max(w, h)
            img = img.resize((int(c*w), int(c*h)))
    return img

def get_image(*args, **kwargs) -> Image:
    print("This function will be removed in future releases. Use load_image() instead.")
    return load_image(*args, **kwargs)

def crop_black(img: Image) -> Image:
    """Crops black borders from an image.    

    Args:
        img (Image): Image to be cropped.

    Returns:
        Cropped image.
    """
    
    y_nonzero, x_nonzero, _ = np.nonzero(img)
    return img.crop(
        (
            np.min(x_nonzero),
            np.min(y_nonzero),
            np.max(x_nonzero),
            np.max(y_nonzero),
        )
    )

def crop_white(img: Image) -> Image:
    """Crops white borders from an image.    

    Args:
        img (Image): Image to be cropped.

    Returns:
        Cropped image.
    """
    
    y_nonzero, x_nonzero, _ = np.nonzero(ImageOps.invert(img))
    return img.crop(
        (
            np.min(x_nonzero),
            np.min(y_nonzero),
            np.max(x_nonzero),
            np.max(y_nonzero),
        )
    )

def find_images(
        root: str,
        img_extensions: Tuple[str, ...] = ('.png', '.jpg', '.jpeg')
        ) -> pd.DataFrame:
    """Finds all image files in folder and subfolders.

    Args:
        root (str): The root folder where to look for images.
        img_extensions (Tuple[str, ...], optional): Image extensions to look for, by default ('.png', '.jpg', '.jpeg').

    Returns:
        Dataframe of relative paths of the images.
    """

    data = [] 
    for path, directories, files in os.walk(root):
        for file in files:
            if file.lower().endswith(tuple(img_extensions)):
                data.append({'path': os.path.relpath(path, start=root), 'file': file})
    return pd.DataFrame(data)

def create_id(string_col: pd.Series) -> pd.Series:
    """Creates unique ids from string based on MD5 hash.

    Args:
        string_col (pd.Series): List of ids.

    Returns:
        List of encoded ids.
    """

    entity_id = string_col.apply(lambda x: hashlib.md5(x.encode()).hexdigest()[:16])
    assert len(entity_id.unique()) == len(entity_id)
    return entity_id

def bbox_segmentation(bbox: List[float]) -> List[float]:
    """Convert bounding box to segmentation.

    Args:
        bbox (List[float]): Bounding box in the form [x, y, w, h].

    Returns:
        Segmentation mask in the form [x1, y1, x2, y2, ...].
    """

    return [bbox[0], bbox[1], bbox[0]+bbox[2], bbox[1], bbox[0]+bbox[2], bbox[1]+bbox[3], bbox[0], bbox[1]+bbox[3], bbox[0], bbox[1]]

def segmentation_bbox(segmentation: List[float]) -> List[float]:
    """Convert segmentation to bounding box.

    Args:
        segmentation (List[float]): Segmentation mask in the form [x1, y1, x2, y2, ...].

    Returns:
        Bounding box in the form [x, y, w, h].
    """

    x = segmentation[0::2]
    y = segmentation[1::2]
    x_min = np.min(x)
    x_max = np.max(x)
    y_min = np.min(y)
    y_max = np.max(y)
    return [x_min, y_min, x_max-x_min, y_max-y_min]

def is_annotation_bbox(
        segmentation: List[float],
        bbox: List[float],
        tol: float = 0
        ) -> bool:
    """Checks whether segmentation is bounding box.

    Args:
        segmentation (List[float]): Segmentation mask in the form [x1, y1, x2, y2, ...].
        bbox (List[float]): Bounding box in the form [x, y, w, h].
        tol (float, optional): Tolerance for difference.

    Returns:
        True if segmentation is bounding box within tolerance.
    """

    bbox_seg = bbox_segmentation(bbox)
    if len(segmentation) == len(bbox_seg):
        for x, y in zip(segmentation, bbox_seg):
            if abs(x-y) > tol:
                return False
    else:
        return False
    return True

class ProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)

def download_url(url, output_path='.'):
    with ProgressBar(unit='B', unit_scale=True, miniters=1, desc=url.split('/')[-1]) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)

def extract_archive(archive, extract_path='.', delete=False):
    shutil.unpack_archive(archive, extract_path)
    if delete:
        os.remove(archive)

@contextmanager
def data_directory(dir):
    '''
    Changes context such that data directory is used as current work directory.
    Data directory is created if it does not exist.
    '''
    current_dir = os.getcwd()
    if not os.path.exists(dir):
        os.makedirs(dir)
    os.chdir(dir)
    try:
        yield
    finally:
        os.chdir(current_dir)

def kaggle_download(command, exception_text='', required_file=None):
    try:
        os.system(f"kaggle {command}")
    except:
        raise Exception(exception_text)
    if required_file is not None and not os.path.exists(required_file):
        raise Exception(exception_text)

def gdown_download(url, archive, exception_text=''):
    import gdown
    gdown.download(url, archive, quiet=False)
    if not os.path.exists(archive):
        print(exception_text)
        raise Exception(exception_text)

def get_split(x, data_train, data_test):
    if x in data_train:
        return 'train'
    elif x in data_test:
        return 'test'
    else:
        return np.nan

def get_image_date(path, shorten=True):
    if not os.path.exists(path):
        return np.nan
    exif = Image.open(path).getexif()
    if exif is not None:
        date = exif.get(36867)
        if date is None: 
            date = exif.get(306)
            if date is not None:
                if len(date) >= 10 and shorten:
                    date = date[:10].replace(':', '-')
                return date
    return np.nan

def yolo_to_pascalvoc(x_c, y_c, w, h, W, H):
    x_min = int((x_c - w / 2) * W)
    y_min = int((y_c - h / 2) * H)
    x_max = int((x_c + w / 2) * W)
    y_max = int((y_c + h / 2) * H)
    return x_min, y_min, x_max, y_max


import os
import datetime
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class MacaqueFaces(DatasetFactory):
    summary = summary['MacaqueFaces']
    
    @classmethod
    def _download(cls):
        downloads = [
            ('https://github.com/clwitham/MacaqueFaces/raw/master/ModelSet/MacaqueFaces.zip', 'MacaqueFaces.zip'),
            ('https://github.com/clwitham/MacaqueFaces/raw/master/ModelSet/MacaqueFaces_ImageInfo.csv', 'MacaqueFaces_ImageInfo.csv'),
        ]
        for url, file in downloads:
            utils.download_url(url, file)

    @classmethod
    def _extract(cls):
        utils.extract_archive('MacaqueFaces.zip', delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Load information about the dataset
        data = pd.read_csv(os.path.join(self.root, 'MacaqueFaces_ImageInfo.csv'))
        date_taken = [datetime.datetime.strptime(date, '%d-%m-%Y').strftime('%Y-%m-%d') for date in data['DateTaken']]
        
        # Finalize the dataframe
        data['Path'] = data['Path'].str.replace('/', os.path.sep)
        df = pd.DataFrame({
            'image_id': pd.Series(range(len(data))),            
            'path': 'MacaqueFaces' + os.path.sep + data['Path'].str.strip(os.path.sep) + os.path.sep + data['FileName'],
            'identity': data['ID'],
            'date': pd.Series(date_taken),
            'category': data['Category']
        })
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class WildlifeReID10k(DatasetFactory):
    summary = summary['WildlifeReID10k']
    archive = 'wildlifereid-10k.zip'

    @classmethod
    def _download(cls):
        command = f"datasets download -d wildlifedatasets/wildlifereid-10k --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/preprocessing#wildlifereid-10k'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        df = pd.read_csv(os.path.join(self.root, 'metadata.csv'), low_memory=False)
        df['image_id'] = range(len(df))
        return self.finalize_catalogue(df)



import os
import json
import numpy as np
import pandas as pd
from . import utils
from .datasets import WildlifeDataset
from .summary import summary

class AnimalCLEF2025(WildlifeDataset):    
    summary = summary['AnimalCLEF2025']
    archive = 'animal-clef-2025.zip'

    @classmethod
    def _download(cls):
        command = f"competitions download -c animal-clef-2025 --force"
        exception_text = '''Kaggle must be setup.
            Check https://wildlifedatasets.github.io/wildlife-datasets/preprocessing#animalclef2025'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        utils.extract_archive(cls.archive, delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        metadata = pd.read_csv(os.path.join(self.root, 'metadata.csv'))
        return self.finalize_catalogue(metadata)

class AnimalCLEF2025_LynxID2025(WildlifeDataset):
    @classmethod
    def _download(cls):
        raise Exception('This dataset is currently available only as part of the AnimalCLEF2025 competition.')

    @classmethod
    def _extract(cls):
        pass
    
    def create_catalogue(self) -> pd.DataFrame:
        data1 = pd.read_csv(os.path.join(self.root, 'metadata_database.csv'))
        data2 = pd.read_csv(os.path.join(self.root, 'metadata_query.csv'))
        data = pd.concat((data1, data2))
        data['path'] = data['path'].str[65:]
        data['species'] = 'lynx'
        data['date'] = np.nan
        return self.finalize_catalogue(data)


class AnimalCLEF2025_SalamanderID2025(WildlifeDataset):
    @classmethod
    def _download(cls):
        raise Exception('This dataset is currently available only as part of the AnimalCLEF2025 competition.')

    @classmethod
    def _extract(cls):
        pass
    
    def create_catalogue(self) -> pd.DataFrame:
        path_json = os.path.join('annotations', 'instances_default.json')

        # Load annotations JSON file
        with open(os.path.join(self.root, path_json)) as file:
            data = json.load(file)

        # Check whether segmentation is different from a box
        for ann in data['annotations']:
            if len(ann['bbox']) != 4:
                raise(Exception('Bounding box missing'))
            if ann['attributes']['rotation'] not in [0, 359.99999999929037]:
                raise(Exception('Rotation is not 0'))

        # Extract the data from the JSON file
        create_dict = lambda i: {'bbox': i['bbox'], 'image_id': i['image_id'], 'orientation': i['attributes']['orientation']}
        df_annotation = pd.DataFrame([create_dict(i) for i in data['annotations']])
        create_dict = lambda i: {'path': i['file_name'], 'image_id': i['id']}
        df_images = pd.DataFrame([create_dict(i) for i in data['images']])

        # Merge the information from the JSON file
        df = pd.merge(df_annotation, df_images, how='left', on='image_id')

        # Include identities
        df['filename'] = df['path'].apply(lambda x: x.split('/')[-1])
        identity = pd.read_csv(os.path.join(self.root, 'metadata.csv'))
        identity = identity[['ts', 'filename', 'identity']]
        df = pd.merge(df, identity, on='filename')
        df = df.drop('filename', axis=1)

        # There are 48 images with two bounding boxes which are actually the same.
        idx = []
        for _, df_file in df.groupby('path'):
            idx.append(df_file.index[0])
        df = df.loc[idx]
        
        # Finalize the dataframe
        df['path'] = df['path'].apply(lambda x: 'images/'+  x.split('/')[-1])
        df = df.rename({'ts': 'date'}, axis=1)
        df['date'] = df['date'].str[:10]
        return self.finalize_catalogue(df)


class AnimalCLEF2025_SeaTurtleID2022(WildlifeDataset):
    @classmethod
    def _download(cls):
        raise Exception('This dataset is currently available only as part of the AnimalCLEF2025 competition.')

    @classmethod
    def _extract(cls):
        pass
    
    def create_catalogue(self) -> pd.DataFrame:
        data = pd.read_csv(os.path.join(self.root, 'annotations.csv'))
        data['image_name'] = data['path'].apply(lambda x: x.split('/')[-1])
        bbox = pd.read_csv(os.path.join(self.root, 'bbox.csv'))
        data = pd.merge(data, bbox, on='image_name')        

        df = pd.DataFrame({
            'image_id': range(len(data)),
            'path': data['path'],
            'identity': data['identity'],
            'date': data['date'],
            'bbox': data[['bbox_x', 'bbox_y', 'bbox_width', 'bbox_height']].values.tolist()
        })
        return self.finalize_catalogue(df)


import os
import numpy as np
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class HumpbackWhaleID(DatasetFactory):
    summary = summary['HumpbackWhaleID']
    archive = 'humpback-whale-identification.zip'

    @classmethod
    def _download(cls):
        command = f"competitions download -c humpback-whale-identification --force"
        exception_text = '''Kaggle terms must be agreed with.
            Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#humpbackwhale'''
        utils.kaggle_download(command, exception_text=exception_text, required_file=cls.archive)

    @classmethod
    def _extract(cls):
        try:
            utils.extract_archive(cls.archive, delete=True)
        except:
            exception_text = '''Extracting failed.
                Either the download was not completed or the Kaggle terms were not agreed with.
                Check https://wildlifedatasets.github.io/wildlife-datasets/downloads#humpbackwhale'''
            raise Exception(exception_text)

    def create_catalogue(self) -> pd.DataFrame:
        # Load the training data
        data = pd.read_csv(os.path.join(self.root, 'train.csv'))
        data.loc[data['Id'] == 'new_whale', 'Id'] = self.unknown_name
        df1 = pd.DataFrame({
            'image_id': data['Image'].str.split('.', expand=True)[0],
            'path': 'train' + os.path.sep + data['Image'],
            'identity': data['Id'],
            'original_split': 'train'
            })
        
        # Find all testing images
        test_files = utils.find_images(os.path.join(self.root, 'test'))
        test_files = list(test_files['file'])
        test_files = pd.Series(np.sort(test_files))

        # Create the testing dataframe
        df2 = pd.DataFrame({
            'image_id': test_files.str.split('.', expand=True)[0],
            'path': 'test' + os.path.sep + test_files,
            'identity': self.unknown_name,
            'original_split': 'test'
            })
        
        # Finalize the dataframe
        df = pd.concat([df1, df2])
        return self.finalize_catalogue(df)



import os
import pandas as pd
from . import utils
from .datasets import DatasetFactory
from .summary import summary

class Drosophila(DatasetFactory):
    summary = summary['Drosophila']
    downloads = [
        ('https://dataverse.scholarsportal.info/api/access/datafile/71066', 'week1_Day1_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71067', 'week1_Day1_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71068', 'week1_Day1_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71069', 'week1_Day1_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71065', 'week1_Day1_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71071', 'week1_Day2_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71072', 'week1_Day2_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71073', 'week1_Day2_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71075', 'week1_Day2_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71070', 'week1_Day2_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71077', 'week1_Day3_01to04.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71078', 'week1_Day3_05to08.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71079', 'week1_Day3_09to12.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71080', 'week1_Day3_13to16.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71081', 'week1_Day3_17to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71083', 'week2_Day1_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71084', 'week2_Day1_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71085', 'week2_Day1_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71086', 'week2_Day1_train_16to20.zip'),        
        ('https://dataverse.scholarsportal.info/api/access/datafile/71082', 'week2_Day1_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71094', 'week2_Day2_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71095', 'week2_Day2_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71109', 'week2_Day2_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71110', 'week2_Day2_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71093', 'week2_Day2_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71111', 'week2_Day3_01to04.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71112', 'week2_Day3_05to08.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71115', 'week2_Day3_09to12.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71117', 'week2_Day3_13to16.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71118', 'week2_Day3_17to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71119', 'week3_Day1_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71120', 'week3_Day1_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71121', 'week3_Day1_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71124', 'week3_Day1_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71097', 'week3_Day1_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71125', 'week3_Day2_train_01to05.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71126', 'week3_Day2_train_06to10.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71127', 'week3_Day2_train_11to15.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71128', 'week3_Day2_train_16to20.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71107', 'week3_Day2_val.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71129', 'week3_Day3_01to04.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71130', 'week3_Day3_05to08.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71131', 'week3_Day3_09to12.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71132', 'week3_Day3_13to16.zip'),
        ('https://dataverse.scholarsportal.info/api/access/datafile/71133', 'week3_Day3_17to20.zip'),
    ]

    @classmethod
    def _download(cls):
        for url, archive in cls.downloads:
            utils.download_url(url, archive)

    @classmethod
    def _extract(cls):
        for url, archive in cls.downloads:
            utils.extract_archive(archive, extract_path=os.path.splitext(archive)[0], delete=True)

    def create_catalogue(self) -> pd.DataFrame:
        # Find all images in root
        data = utils.find_images(self.root)
        folders = data['path'].str.split(os.path.sep, expand=True)

        # Extract information from the folder structure
        data['identity'] = self.unknown_name
        for i_week in range(1, 4):
            idx1 = folders[0].str.startswith('week' + str(i_week))
            idx2 = folders[1] == 'val'
            idx3 = folders[2].isnull()
            data.loc[idx1 & ~idx2, 'identity'] = ((i_week-1)*20 + folders.loc[idx1 & ~idx2, 1].astype(int)).astype(str)
            data.loc[idx1 & idx2 & ~idx3, 'identity'] = ((i_week-1)*20 + folders.loc[idx1 & idx2 & ~idx3, 2].astype(int)).astype(str)
            data.loc[idx1 & ~idx2, 'split'] = 'train'
            data.loc[idx1 & idx2, 'split'] = 'val'
        
        # Create id and path
        data['id'] = utils.create_id(folders[0] + data['file'])
        data['path'] = data['path'] + os.path.sep + data['file']
        
        # Finalize the dataframe
        df = data.drop(['file'], axis=1)
        df.rename({'id': 'image_id'}, axis=1, inplace=True)
        return self.finalize_catalogue(df)
